
// *******************************************************************************************************
// *******************************************************************************************************
// ************
// ************ Project:      Basic Profinet Example for communication with a controller - KUNBUS GmbH
// ************
// ************ Subject:      Communication with a controller board by use of dfpnio.dll functions
// ************ Filename:     Basic_Profinet_Example.c
// ************ Company:      KUNBUS GmbH, Robert-Bosch-Strasse 12a, D-76275 Ettlingen
// ************ Autor:        CH, 11.07.2013
// ************ Description:  This software is a BASIC example how to use a Profinet controller board.
// ************
// ************ Expanded:     v1.1  12.07.2013 ch  : additional functions and alarm handling
// ************               v1.2  21.10.2013 SF  : ModuleDiffBlock added
// ************               v1.3  22.10.2013 SF  : RW RecordData added
// ************               v1.4  15.01.2014 BS  : Improvements for the screen output.
// ************               v1.5  16.01.2015 CVO : Bugfixes and screen output improvements
// ************               v1.6  22.07.2015 CVO : Added Single Steps into Menu
// ************               v1.7  23.09.2015 CVO : Added latest dfpnio.dll
// ************               v1.8  23.09.2015 CVO : Changed DLL function names
// ************               v1.9  14.07.2016 CVO : Change to KUNBUS, new DLL
// ************
// *******************************************************************************************************
// *******************************************************************************************************
#define          VERSION_STR "v1.9"

//   What this application wants to demonstrate
//
// - Use this software parallel to the documention about the Profinet Controller.
// - This software is a BASIC example how to use a Profinet Controller with dfpnio.dll
// - The intention is to give a short example how the Profinet communication is working.
//   Only the absolut basic functionality is used.
// - Use the Microsoft Visual Studio Debugger to singlestep this software. Without steping it will be harder
//   to understand how the interface works.
//
//
//   Requirements for compiling and running this demo successfull on your system
//
// - Windows XP or higher
// - Microsoft Visual Studio 2008. Other compiler will also work but are not tested. Maybe some errors could occure.
// - The setup for the Profinet controller driver has to be executed successfully.
// - At least 1 Profinet controller hardware is placed in the PC.
// - Use the KUNBUS "Configurator 3" to generate a project configuration XML file and download it
// - To use the Profinet controller is it useful to connect a profinet device to communicate with.
//
//
//   How the demo works
//
//   There are 10 steps
//   Step 1: optional: Check the DLL version of dfpnio.dll
//   Step 2: Get information about the number of installed controllers, names of the controller and
//           information about where inthe PC (PCI slot) the board is placed.
//   Step 3: Choose a controller number from the installed hardware.
//   Step 4: optional: Get information about all devices in the Profinet by a DCP call IdentifyAll.
//   Step 5: Start Profinet.
//   Step 6: Select device
//   Step 7: Poll APDU status of the selected device until status is ok or timeout reached.
//   Step 8: Set output data and get current input data from the selected device and check the device status.
//           In case of any alarm on this device, show the alarm information and alarm data.
//           In case of ModuleDiffBlock show the information.
//   Step 9: Read/Write RecordData with selected device
//   Step 10: Stop Profinet.
//
//   Hints and tips for programming:
// - If a new Visual Studio project is created, 1 external library dfpnio.lib has to be added to the
//   linkers library list.
//   For Visual Studio 2008: Go to project/settings/All Configurations/Linker/Object-LibraryModules/
//   there you have to add the library dfpnio.lib (separated with a blank, if other libs are already included).
// - pointer has a small "p" as the first letter in their name (for example "pControllerInfo" is a pointer,
//   "DeviceInfo" is an instance)


// **************** Standard C Header includes ***********************************************************

#ifdef __linux__
	#include <unistd.h>
	#include <string.h>
	#define _getch() getchar()
	#define Sleep(ms) usleep(ms*1000)
	#define scanf_s scanf
	#define sprintf_s(a,b,...) sprintf(a, __VA_ARGS__)
#else
	#include <Windows.h>
	#include <conio.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>


// **************** Header includes for Profinet Communication under Windows *******************************
#include "dfpnio.h"
#include "PN_LibMain.h"
#include "comm_interface_errors.h"



//*********************** Prototypes of used subfunctions ******************
void Step1_GetDLLVersion(void);
void Step2_GetControllerInfo(void);
void Step3_ChooseController(void);
void Step4_IdentifyAllDevices(void);
void Step5_StartProfinet(void);
void Step6_SelectDevice(void);
void Step7_GetAPDUStatus(INT32U device_nr);
void Step8_SetGetDataWithAlarmHandling(INT32U device_nr);
void Step9_ReadWriteRecordDataHandling(INT32U device_nr);
void Step10_StopProfinet(void);

void AlarmHandling(INT32U  device_nr, INT8U DeviceStatus);
void ModuleDiffBockHandling(INT32U  device_nr);
void PrintSubmoduleState(USIGN16 wSubmoduleState);
void PrintModuleState(USIGN16 wModuleState);


//*********************** Global variables ******************
/*
   error_class: in case of a command was not processed correctly. In case of a
   error_class != 0 additional information is found in the error value,
   which can be read by a call to GetControllerError(board_nr).
   Or full analyzed error status in text form by a call to
   GetDLLErrorText(error_class, GetControllerError(board_nr))
*/
   INT32U error_class = ERROR_CLASS_NO_ERROR;  // return value of every DLL function
   INT32U board_count = 0;                // number of Profinet controllers in the system
   INT32U board_nr    = 0;                // controller to be used in this program
   INT32U device_nr  = 0;                // device to be used in this program


static const CHAR *pszModuleState[] =
{
    "MODULE_STATE_NO_MODULE",
    "MODULE_STATE_WRONG_MODULE",
    "MODULE_STATE_PROPER_MODULE",
    "MODULE_STATE_SUBSTITUTE"
};

static const CHAR *pszSubmoduleState[] =
{
    "SUBMODULE_STATE_ADD_INFO_NONE",
    "SUBMODULE_STATE_ADD_INFO_TAKEOVER_NOT_ALLOWED",
    "SUBMODULE_STATE_RESEREVED",
    "SUBMODULE_STATE_MAINTENANCE_REQUIRED_AVAILABLE",
    "SUBMODULE_STATE_MAINTENANCE_DEMANDED_AVAILABLE",
    "SUBMODULE_STATE_FAULT",
    "SUBMODULE_STATE_AR_INFO_OWN",
    "SUBMODULE_STATE_AR_INFO_APPLICATION_READY_PENDING",
    "SUBMODULE_STATE_AR_INFO_SUPERORDINATION_LOCKED",
    "SUBMODULE_STATE_AR_INFO_LOCKED_BY_IO_CONTROLLER",
    "SUBMODULE_STATE_AR_INFO_LOCKED_BY_IO_SUPERVISOR",
    "SUBMODULE_STATE_IDENT_INFO_OK",
    "SUBMODULE_STATE_IDENT_INFO_SUBSTITUTE",
    "SUBMODULE_STATE_IDENT_INFO_WRONG",
    "SUBMODULE_STATE_IDENT_INFO_NO_SUBMODULE"
};

static const CHAR *pszSubmoduleStateDetail[] =
{
    "SUBMODULE_STATE_DETAIL_NO_SUBMODULE",
    "SUBMODULE_STATE_DETAIL_WRONG_SUBMODULE",
    "SUBMODULE_STATE_DETAIL_LOCKED_BY_IO_CONTROLLER",
    "SUBMODULE_STATE_DETAIL_APPLICATION_READY_PENDING",
    "SUBMODULE_STATE_DETAIL_SUBSTITUTE"
};

//*************************************** MAIN FUNCTION **************************************************
int main( void )
{
   int Mode = 0;

   printf("\n\n\nBasic Profinet Example %s - KUNBUS GmbH\n\n", VERSION_STR);

   //Step 1: optional: Check DLL version
   Step1_GetDLLVersion();

   //Step 2: Get information about the number of installed controlers, names and installation PCI slot
   Step2_GetControllerInfo();

   //Step 3: Choose a controller
   Step3_ChooseController();

   do {
      printf("\n\nCommand");
      printf("\n 0 - Exit");
      printf("\n 1 - Identify Devices on Bus");
      printf("\n 2 - Cyclic Transfer");
      printf("\n 3 - Read/Write Records");
      printf("\n     ----- Single Steps -------");
      printf("\n 4 - Start Profinet");
      printf("\n 5 - Stop Profinet");
      printf("\n 6 - Select Device");
      printf("\n 7 - Get APDU Status");
      printf("\n 8 - Set/Get Cyclic Data");
      printf("\nSelect Command: ");
      scanf_s("%d", &Mode);

      switch (Mode)
      {
      case 0:
         // exit
         break;

      case 1:
         //Step 4: optional: Get information about all devices by a DCP call IdentifyAll
         Step4_IdentifyAllDevices();
         break;

      case 2:
         //Step 5: Start Profinet
         Step5_StartProfinet();

         //Step 6: Select Device
         Step6_SelectDevice();

         //Step 7: Poll APDU status of the device 1 until status is ok or timeout reached
         Step7_GetAPDUStatus(device_nr);

         //Step 8: Set output data and get input data from the device 1 and check the device status
         Step8_SetGetDataWithAlarmHandling(device_nr);

         //Step 10: Stop Profinet
         Step10_StopProfinet();
         break;

      case 3:
         //Step 5: Start Profinet
         Step5_StartProfinet();

         //Step 6: Select Device
         Step6_SelectDevice();

         //Step 7: Poll APDU status of the device 1 until status is ok or timeout reached
         Step7_GetAPDUStatus(device_nr);

         //Step 9:
         Step9_ReadWriteRecordDataHandling(device_nr);

         //Step 10: Stop Profinet
         Step10_StopProfinet();
         break;


      case 4:
         //Step 5: Start Profinet
         Step5_StartProfinet();

         break;

      case 5:
         //Step 10: Stop Profinet
         Step10_StopProfinet();

         break;

      case 6:
         //Step 6: Select Device
         Step6_SelectDevice();

         break;

      case 7:
         //Step 7: Poll APDU status of the device 1 until status is ok or timeout reached
         Step7_GetAPDUStatus(device_nr);

         break;

      case 8:
         //Step 8: Set output data and get input data from the device 1 and check the device status
         Step8_SetGetDataWithAlarmHandling(device_nr);

         break;

      default:
         printf("\n Invalid Input \n");
         break;
      }
   }while(Mode != 0);

   printf("\n\n\nFinished Basic Profinet Example. Press key ...." );
   _getch();
   return(0);
}



//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step1_GetDLLVersion(void)
{
   printf("\n\nStep 1 (optional): Check DLL version. Press key...");
   _getch();
   {
      char* pversion_string;
      // DLL function call; the string is managed by the DLL
      CsPnioGetDllVersion(&pversion_string); // return value not used, because this DLL function is always successful
      printf("\nDLL Version used is : %s", pversion_string);
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step2_GetControllerInfo(void)
{
   printf("\n\nStep 2: Get controller info. Press key...");
   _getch();
   {
      CONTROLLER_INFO* pControllerInfo;
      // DLL function call; the memory of the controller info structure is managed by the DLL
      error_class = CsPnioGetControllerInfo(&board_count, &pControllerInfo);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         INT32U i;
         for (i = 0; i < board_count; i++)
         {
            printf("\nController # %d  named: %s  placed in PCI bus/device/function %d",
               pControllerInfo->Index, pControllerInfo->Name, pControllerInfo->PCIslot);
            pControllerInfo++;
         }
      }
      else
      {
         printf("\nController Init failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(0)));
      }
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step3_ChooseController(void)
{
   printf("\n\nStep 3: Choose a controller. Press key...");
   _getch();
   // the board_count is known after step 2
   if (board_count > 0)
   {
      INT32U KeybordInput;
      do
      {
         printf("\r\nEnter controller number 0 to %d: ", board_count - 1);
         KeybordInput = scanf_s("%d", &board_nr);
         if (KeybordInput  == 0) //empty keyboardbuffer if value was not readable
         {
            scanf_s("%u", &KeybordInput);
            KeybordInput=0;
         }
      }while ((KeybordInput == 0) || (board_nr > board_count - 1));
      printf("Controller # selected: %d", board_nr);
   }
   else
   {
      printf("\nNo controller installed or driver not started");
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step4_IdentifyAllDevices(void)
{
   printf("\n\nStep 4: Get information about all devices. Press key...");
   _getch();
   {
      INT32U DcpSearchResultEntrys;
      INT32U DcpSearchSize;
      DFPN_DCP_RESULT* pParam;

      // the command blocks the application until the IdentifyAll is done or timed out (aprox. 10s)
      printf("\nDCP identify in progress. Wait 10s...");
      // DLL function call; the DCP structure is managed by the DLL
      error_class = CsPnioIdentifyAllDevices(board_nr, &DcpSearchResultEntrys, &DcpSearchSize, (char**)&pParam);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         INT32U i;
         printf("\nDCP identify found %d device(s):", DcpSearchResultEntrys);

         for (i = 0; i < DcpSearchResultEntrys; i++)
         {
            printf("\n%u. %s", i+1, pParam[i].StationName);
         }
      }
      else
      {
         printf("\nDCP_IdentifyAll failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step5_StartProfinet(void)
{
   printf("\n\nStep 5: Start the Profinet. Press key...");
   _getch();
   {
      // DLL function call; the function initializes some DLL internal structures and starts the Profinet
      error_class = CsPnioInitStartProfinet(board_nr);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nProfinet start Controller %d OK", board_nr);
      }
      else
      {
         printf("\nProfinet Controller %d not started. %s", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}

//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step6_SelectDevice(void)
{
   // the board_count is known after step 2
   INT32U KeybordInput;
   int i;
   DEVICE_INFO DeviceInfo;

   printf("\n\nStep 6: Select device number.");

   // Scan for available devices.
   // this is not necessary if the configuratin is already known
   printf("\nAvailable Devices: ");
   for (i=1; i<128; i++)
   {
      error_class = CsPnioGetDeviceData(board_nr, i, &DeviceInfo);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("%d, ", i);
      }
      else if (error_class == ERROR_CLASS_DLL && CsPnioGetControllerError(board_nr) == RET_DLL_DEVICE_INDEX_INVALID)
      {
         // this is ok, because we are searching for valid index.
      }
      else
      {
         printf("\nGetDeviceData failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
   printf("\n");

   do
   {
      printf("\r\nEnter device number: ");
      KeybordInput = scanf_s("%u", &device_nr);
      if (KeybordInput  == 0) //empty keyboardbuffer if value was not readable
      {
         scanf_s("%u", &KeybordInput);
         KeybordInput=0;
      }
   }while ((KeybordInput == 0) || (device_nr < 1) || (device_nr > 127));
   printf("Device # selected: %u", device_nr);
}

//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step7_GetAPDUStatus(INT32U device_nr)
{
   printf("\n\nStep 7: Get APDU status of device %u. Press key...", device_nr);
   _getch();
   {
      // the device info structure is filled in the DLL.
      DEVICE_INFO DeviceInfo;
      // The status defines are from PN_LibMain.h
      INT8U APDUstatusOK = DS_STATION_NO_PROBLEM | DS_PROV_STATE_RUN | DS_DATA_VALID | DS_STATE_PRIMARY;
      // wait 20 loops of 500 ms = 10s time for the device to reach APDU status ok
      INT16U timeout = 20;
      // The APDU status is polled until ok or timeout
      printf("\n");
      do {
         timeout--;
         // DLL function call; the device info structure is filled in the DLL.
         error_class = CsPnioGetDeviceData(board_nr, device_nr, &DeviceInfo);
         if (error_class == ERROR_CLASS_NO_ERROR)
         {
            if (DeviceInfo.APDUStatus != APDUstatusOK )
            {
               printf("0x%X  ", DeviceInfo.APDUStatus);
               Sleep(500);
               // update device data in dll internal structures for next GetDevicedata
               error_class = CsPnioRW_ProcessData(board_nr);
               if (error_class == ERROR_CLASS_NO_ERROR)
               {
               }
               else
               {
                  printf("\nProcess image access failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
                  break;
               }
            }
         }
         else
         {
            printf("\nGet device data failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
            break;
         }
      } while (DeviceInfo.APDUStatus != APDUstatusOK && timeout > 0);

      if (DeviceInfo.APDUStatus == APDUstatusOK)
      {
         printf("\nAPDU Status OK: 0x%X", DeviceInfo.APDUStatus);
      }
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step8_SetGetDataWithAlarmHandling(INT32U device_nr)
{
   {
      // 10 times Set output data and get input data from the device and check the device status
      // if the device status indicates a problem, then get the diagnose data
      int i;
      INT8U data_byte;
      DEVICE_INFO DeviceInfo;

      data_byte = 0x77;

      // the first call of GetDeviceData is used to initalize DeviceInfo.
      // We will get pointer and length information to write outputs
      error_class = CsPnioGetDeviceData(board_nr, device_nr, &DeviceInfo);
      if (error_class != ERROR_CLASS_NO_ERROR)
      {
         printf("\nGetDeviceData failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
         return;
      }

      for (i = 1; i <= 10; i++)
      {
         printf("\n\n\n\nStep 8.%u: Set output data and get input data %u out of 10. Press key...",i,i);
         _getch();

         // Set Output Data of the current device
         DeviceInfo.OutputsStatus = GOOD;  // Output Provider Status is GOOD
         memset(DeviceInfo.pOutputData, data_byte + i, DeviceInfo.OutputsLength);

         // DLL function call; the data are written to DLL internal structures, not yet to the controller
         error_class = CsPnioSetDeviceData(board_nr, device_nr, &DeviceInfo);
         if (error_class == ERROR_CLASS_NO_ERROR)
         {
            // refresh the process image: write the output values to the controller and receive the current input data
            error_class = CsPnioRW_ProcessData(board_nr);
            if (error_class == ERROR_CLASS_NO_ERROR)
            {
               // read the current input data into DLL structures where DeviceInfo.pInputData reads from
               error_class = CsPnioGetDeviceData(board_nr, device_nr, &DeviceInfo);
               if (error_class == ERROR_CLASS_NO_ERROR)
               {
                  // check for problems
                  if (DeviceInfo.DeviceStatus & DFPN_DIAGNOSIS_ALARM)
                  {
                     printf("\n");
                     AlarmHandling(device_nr, DFPN_DIAGNOSIS_ALARM);
                  }
                  if (DeviceInfo.DeviceStatus & DFPN_PROCESS_ALARM)
                  {
                     printf("\n");
                     AlarmHandling(device_nr, DFPN_PROCESS_ALARM);
                  }
				      if (DeviceInfo.DeviceStatus & DFPN_MODULE_DIFF_BLOCK_FLAG)
				      {
                     printf("\n");
					      ModuleDiffBockHandling(device_nr);
				      }

                  // show data content
                  if ((DeviceInfo.APDUStatus & DS_DATA_VALID) == 0)
                  {
                     printf("\nProcess Data is not valid. Maybe connection is lost. ");
                     printf("APDU Status: 0x%X", DeviceInfo.APDUStatus);
                  }
                  else
                  {
                     INT16U i;
                     printf("\n\nDeviceNo %u, APDU 0x%X, Device status 0x%X, Input length %d, sum status 0x%X",
                        device_nr, DeviceInfo.APDUStatus, DeviceInfo.DeviceStatus, DeviceInfo.InputsLength, DeviceInfo.InputsStatus);

                     if (DeviceInfo.InputsStatus == GOOD)
                        printf(" (GOOD)");
                     else
                        printf(" (BAD)");

                     printf("\n\nData In (Hex) = ");
                     for (i = 0; i < DeviceInfo.InputsLength; i++)
                     {
                        printf("%02X ", DeviceInfo.pInputData[i]);
                        if ( !((i+1) % 16 ))
                        {
                           printf("\n                 ");
                        }
                     }

                     printf("\n\nData Out (Hex)  = ");
                     for (i = 0; i < DeviceInfo.OutputsLength; i++)
                     {
                        printf("%02X ", DeviceInfo.pOutputData[i]);
                        if ( !((i+1) % 16 ))
                        {
                           printf("\n                 ");
                        }
                     }
                  }
               }
               else
               {
                  printf("\nGet device data failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
                  break;
               }
            }
            else
            {
               printf("\nProcess image access failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
               break;
            }
         }
         else
         {
            printf("\nSet device data failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
            break;
         }
      }
   }
}

//***********************************************************************************************************************
void AlarmHandling(INT32U  device_nr, INT8U DeviceStatus)
{
   // get alarm information from the device.
   // acknowledge alarm is done always by auto-acknowledge on the controller
   // show alarm class (process or diagnose alarm) and alarm data

   INT32U AlarmClass;

   if (DeviceStatus & DFPN_DIAGNOSIS_ALARM)
   {
      AlarmClass = DFPN_DIAGNOSIS_ALARM;
   }
   else if (DeviceStatus & DFPN_PROCESS_ALARM)
   {
      AlarmClass = DFPN_PROCESS_ALARM;
   }
   else // invalid DeviceStatus (normally not possible)
   {
      AlarmClass = 0;
   }

   if (AlarmClass != 0)
   {
      DFPN_ALARM_NOTIFICATION AlarmNotification;

      error_class = CsPnioAlarmGet(board_nr, device_nr, AlarmClass, &AlarmNotification);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         INT32U i;
          INT32U len = AlarmNotification.LengthOfAlarmItem;

         if (AlarmClass == DFPN_DIAGNOSIS_ALARM)
         {
            printf("\nDiagnose alarm data on device %u", device_nr);
         }
         else
         {
            printf("\nProcess alarm data on device %u", device_nr);
         }
         printf("\nSlot                %d", AlarmNotification.SlotNumber);
         printf("\nSubslot             %d", AlarmNotification.SubslotNumber);
         printf("\nModule Ident        %u", AlarmNotification.ModuleIdentNumber);
         printf("\nSubmodule Ident     %u", AlarmNotification.SubmoduleIdentNumber);
         printf("\nBlockType           %d", AlarmNotification.BlockType);
         printf("\nBlockVersion:       %d", AlarmNotification.BlockVersion);
         printf("\nBlock Length        %d", AlarmNotification.BlockLength);
         printf("\nAlarm Type          %d", AlarmNotification.AlarmType);
         printf("\nAPI identifier      %u", AlarmNotification.API);
         printf("\nAlarm Specifier:    %d", AlarmNotification.AlarmSpecifier);
         printf("\nAlarm Item Length:  %d\n", AlarmNotification.LengthOfAlarmItem);
         for (i = 0; i < len; i++)
         {
            printf("0x%X ", AlarmNotification.pAlarmItem[i]);
            if ( !((i+1) % 16 ))
            {
               printf("\n");
            }
         }
      }
      else
      {
         printf("\nGetAlarm class %u failed. %s", AlarmClass, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}

//***********************************************************************************************************************
void ModuleDiffBockHandling(INT32U device_nr)
{
   // get ModuleDiffBlock information from the device.
   DFPN_MODULE_DIFF_BLOCK *ModuleDiffBlock;

   error_class = CsPnioModuleDiffBlockGet(board_nr, device_nr, &ModuleDiffBlock);
   if (error_class == ERROR_CLASS_NO_ERROR)
   {
      INT32U i, j;

      printf("\nModuleDiffBlock on device %u", device_nr);
      printf("\nARUID             : %u", device_nr);
      printf("\nARID              : %u", ModuleDiffBlock->ARID);
      printf("\nStationName       : %s", ModuleDiffBlock->StationName);
      printf("\nNumberOfModules   : %u", ModuleDiffBlock->NumberOfModules);
      for (i = 0; i < ModuleDiffBlock->NumberOfModules; i++)
      {
         printf("\n   Module             : %i", i);
         printf("\n   SlotNumber         : %u", ModuleDiffBlock->ListOfModules[i].SlotNumber);
         printf("\n   ModuleIdentNumber  : 0x%04x", ModuleDiffBlock->ListOfModules[i].ModuleIdentNumber);
         printf("\n   ModuleState        : 0x%04x", ModuleDiffBlock->ListOfModules[i].ModuleState);
         PrintModuleState((INT16U)ModuleDiffBlock->ListOfModules[i].ModuleState);
         printf("\n   NumberOfSubmodules : %u", ModuleDiffBlock->ListOfModules[i].NumberOfSubmodules);
         for (j = 0; j < ModuleDiffBlock->ListOfModules[i].NumberOfSubmodules; j++)
         {
            printf("\n      Submodule         : %u", j);
            printf("\n      SubslotNumber     : %u", ModuleDiffBlock->ListOfModules[i].ListOfSubmodules[j].SubslotNumber);
            printf("\n      SubmoduleIdentNum : 0x%04x", ModuleDiffBlock->ListOfModules[i].ListOfSubmodules[j].SubmoduleIdentNumber);
            printf("\n      SubmoduleState    : 0x%04x", ModuleDiffBlock->ListOfModules[i].ListOfSubmodules[j].SubmoduleState);
            PrintSubmoduleState((INT16U)ModuleDiffBlock->ListOfModules[i].ListOfSubmodules[j].SubmoduleState);
         }
      }
   }
   else
   {
      printf("\nModuleDiffBockHandling failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
   }
}

//***********************************************************************************************************************
void PrintModuleState(INT16U wModuleState)
{
    char szTabulator[] = {"                        "};

    switch (wModuleState)
    {
        case MODULE_STATE_NO_MODULE:
            /* E.g. module not plugged */
            printf("\n%s%s", szTabulator, pszModuleState[0]);
            break;
        case MODULE_STATE_WRONG_MODULE:
            /* E.g. module ident number wrong */
            printf("\n%s%s", szTabulator, pszModuleState[1]);
            break;
        case MODULE_STATE_PROPER_MODULE:
            /* Module is okay but at least one submodule is locked, wrong or missing */
            printf("\n%s%s", szTabulator, pszModuleState[2]);
            break;
        case MODULE_STATE_SUBSTITUTE:
            /* Module is not the same as requested � but compatible.
               The IO device was able to adapt by its own decision   */
            printf("\n%s%s", szTabulator, pszModuleState[3]);
            break;
    }
}

//***********************************************************************************************************************
void PrintSubmoduleState(INT16U wSubmoduleState)
{
    char szTabulator[] = {"                          "};
    INT16U wMaskedValue = 0;

    /* Bit 15: SubmoduleState.FormatIndicator */
    if ((wSubmoduleState & 0x8000) == 0x8000)
    {
        /* Coding uses SubmoduleState.IdentInfo, SubmoduleState.ARInfo and SubmoduleState.AddInfo */

        /* Bit 0 � 2: SubmoduleState.AddInfo */
        wMaskedValue = wSubmoduleState & SUBMODULE_STATE_ADD_INFO_MASK;
        switch (wMaskedValue)
        {
            case SUBMODULE_STATE_ADD_INFO_NONE:
                printf("\n%s%s", szTabulator, pszSubmoduleState[0]);
                break;
            case SUBMODULE_STATE_ADD_INFO_TAKEOVER_NOT_ALLOWED:
                /* This Submodul is not available for takeover by IOSAR. */
                printf("\n%s%s", szTabulator, pszSubmoduleState[1]);
                break;
        }

        /* Bit 3: SubmoduleState.Reserved */
        if (wSubmoduleState & SUBMODULE_STATE_RESERVED)
        {
            printf("\n%s%s", szTabulator, pszSubmoduleState[2]);
        }

        /* Bit 4: SubmoduleState.MaintenanceRequired */
        if (wSubmoduleState & SUBMODULE_STATE_MAINTENANCE_REQUIRED_AVAILABLE)
        {
            /* At least one Channel of the Submodule requires maintenance */
            printf("\n%s%s", szTabulator, pszSubmoduleState[3]);
        }

        /* Bit 5: SubmoduleState.MaintenanceDemanded */
        if (wSubmoduleState & SUBMODULE_STATE_MAINTENANCE_DEMANDED_AVAILABLE)
        {
            /* At least one Channel of the Submodule demandes maintenance */
            printf("\n%s%s", szTabulator, pszSubmoduleState[4]);
        }

        /* Bit 6: SubmoduleState.StateFault */
        if (wSubmoduleState & SUBMODULE_STATE_FAULT)
        {
            /* There is DiagnosisData available for this submodule:
               It can be read with the corresponding records. */
            printf("\n%s%s", szTabulator, pszSubmoduleState[5]);
        }

        /* Bit 7 � 10: SubmoduleState.ARInfo */
        wMaskedValue = wSubmoduleState & SUBMODULE_STATE_AR_INFO_MASK;
        switch (wMaskedValue)
        {
            case SUBMODULE_STATE_AR_INFO_OWN:
                /* This AR is owner of the submodule */
                printf("\n%s%s", szTabulator, pszSubmoduleState[6]);
                break;
            case SUBMODULE_STATE_AR_INFO_APPLICATION_READY_PENDING:
                /* This AR is owner of the submodule but it is blocked,
                   e.g. parameter checking pending */
                printf("\n%s%s", szTabulator, pszSubmoduleState[7]);
                break;
            case SUBMODULE_STATE_AR_INFO_SUPERORDINATION_LOCKED:
                /* This AR is not owner of the submodule.
                   It is blocked by superordinated means */
                printf("\n%s%s", szTabulator, pszSubmoduleState[8]);
                break;
            case SUBMODULE_STATE_AR_INFO_LOCKED_BY_IO_CONTROLLER:
                /* This AR is not owner of the submodule.
                   It is owned by an other IOAR */
                printf("\n%s%s", szTabulator, pszSubmoduleState[9]);
                break;
            case SUBMODULE_STATE_AR_INFO_LOCKED_BY_IO_SUPERVISOR:
                /* This AR is not owner of the submodule.
                   It is owned by an other IOSAR */
                printf("\n%s%s", szTabulator, pszSubmoduleState[10]);
                break;
        }

        /* Bit 11 � 14: SubmoduleState.IdentInfo */
        wMaskedValue = wSubmoduleState & SUBMODULE_STATE_IDENT_INFO_MASK;
        switch (wMaskedValue)
        {
            case SUBMODULE_STATE_IDENT_INFO_OK:
                printf("\n%s%s", szTabulator, pszSubmoduleState[11]);
                break;
            case SUBMODULE_STATE_IDENT_INFO_SUBSTITUTE:
                printf("\n%s%s", szTabulator, pszSubmoduleState[12]);
                break;
            case SUBMODULE_STATE_IDENT_INFO_WRONG:
                printf("\n%s%s", szTabulator, pszSubmoduleState[13]);
                break;
            case SUBMODULE_STATE_IDENT_INFO_NO_SUBMODULE:
                printf("\n%s%s", szTabulator, pszSubmoduleState[14]);
                break;
        }
    }
    else
    {
        /* Coding uses SubmoduleState.Detail */

        /* Bit 0 � 14: SubmoduleState.Detail */
        wMaskedValue = wSubmoduleState & SUBMODULE_STATE_DETAIL_MASK;
        switch (wMaskedValue)
        {
            case SUBMODULE_STATE_DETAIL_NO_SUBMODULE:
                printf("\n%s%s", szTabulator, pszSubmoduleStateDetail[0]);
                break;
            case SUBMODULE_STATE_DETAIL_WRONG_SUBMODULE:
                printf("\n%s%s", szTabulator, pszSubmoduleStateDetail[1]);
                break;
            case SUBMODULE_STATE_DETAIL_LOCKED_BY_IO_CONTROLLER:
                printf("\n%s%s", szTabulator, pszSubmoduleStateDetail[2]);
                break;
            case SUBMODULE_STATE_DETAIL_APPLICATION_READY_PENDING:
                printf("\n%s%s", szTabulator, pszSubmoduleStateDetail[3]);
                break;
            case SUBMODULE_STATE_DETAIL_SUBSTITUTE:
                printf("\n%s%s", szTabulator, pszSubmoduleStateDetail[3]);
                break;
        }
    }
}

//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step9_ReadWriteRecordDataHandling(INT32U device_nr)
{
   INT32U i;
   printf("\n\nStep 9: Read / Write Record Data. Press key...");
   _getch();
   // Read RecordData
   {

      DFPN_RECORD_RESULT ReadResult;
      INT16U SlotNumber = 1;
      INT16U SubslotNumber = 1;
      INT16U Index = 0x2010;
      INT8U  Data[256];
      INT16U DataLength = sizeof(Data);
      int            ReadValid  = 0;

      printf("\nRead Record Data on device %u", device_nr);
      printf("\nSlotNumber        = %d", SlotNumber);
      printf("\nSubslotNumber     = %d", SubslotNumber);
      printf("\nIndex             = 0x%X", Index);
      printf("\nSlot              = %d", SlotNumber);
      printf("\nDataLength        = %d", DataLength);

      error_class = CsPnioReadRecord(board_nr, device_nr, SlotNumber, SubslotNumber, Index, Data, DataLength, &ReadResult);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nReturnValues");

         if (ReadResult.ErrorCode == PNIO_STATUS_NO_ERROR && ReadResult.ErrorDecode == PNIO_STATUS_NO_ERROR
            && ReadResult.ErrorCode1 == PNIO_STATUS_NO_ERROR)
         {
            printf("\n Read successful");
            ReadValid = 1;
         }
         else
         {
            printf("\n Read with error");
            ReadValid = 0;
         }

         printf("\nErrorCode         = 0x%X", ReadResult.ErrorCode);
         printf("\nErrorDecode       = 0x%X", ReadResult.ErrorDecode);
         printf("\nErrorCode1        = 0x%X", ReadResult.ErrorCode1);
         printf("\nErrorCode2        = 0x%X", ReadResult.ErrorCode2);
         printf("\nAddData1          = 0x%X", ReadResult.AddData1);
         printf("\nAddData2          = 0x%X", ReadResult.AddData2);
         printf("\nDataLength        = %u",   ReadResult.DataLength);

         if (ReadValid)
         {
            printf("\nData (Hex)        = ");

            for (i = 0; i < ReadResult.DataLength; i++)
            {
               CHAR temp[10];
               sprintf_s(temp, sizeof(temp), "%02X ", Data[i]);
               printf("%s", temp);
               if ( !((i+1) % 16 ))
               {
                  printf("\n                    ");
               }
            }
         }
         printf("\n");
      }
      else
      {
         printf("\nRead RecordData failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }

   // Write RecordData
   {
      DFPN_RECORD_RESULT WriteResult;
      INT16U SlotNumber = 1;
      INT16U SubslotNumber = 1;
      INT16U Index = 0x2011;
      INT8U  Data[4];
      INT16U DataLength = sizeof(Data);

      printf("\nWrite Record Data on device %u", device_nr);

      Data[0] = 0xAF;
      Data[1] = 0xFE;
      Data[2] = 0x11;
      Data[3] = 0x22;

      printf("\nSlotNumber        = %d", SlotNumber);
      printf("\nSubslotNumber     = %d", SubslotNumber);
      printf("\nIndex             = 0x%X", Index);
      printf("\nSlot              = %d", SlotNumber);
      printf("\nDataLength        = %d", DataLength);
      printf("\nData (Hex)        = ");
      for (i = 0; i < DataLength; i++)
      {
         CHAR temp[10];
         sprintf_s(temp, sizeof(temp), "%02X ", Data[i]);
         printf("%s", temp);
         if ( !((i+1) % 16 ))
         {
            printf("\n                   ");
         }
      }
      printf("\n");

      error_class = CsPnioWriteRecord(board_nr, device_nr, SlotNumber, SubslotNumber, Index, Data, DataLength, &WriteResult);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nReturnValues");

         if (WriteResult.ErrorCode == PNIO_STATUS_NO_ERROR && WriteResult.ErrorDecode == PNIO_STATUS_NO_ERROR
            && WriteResult.ErrorCode1 == PNIO_STATUS_NO_ERROR)
         {
            printf("\n Write successful");
         }
         else
         {
            printf("\n Write with error");
         }

         printf("\nErrorCode        = 0x%X", WriteResult.ErrorCode);
         printf("\nErrorDecode      = 0x%X", WriteResult.ErrorDecode);
         printf("\nErrorCode1       = 0x%X", WriteResult.ErrorCode1);
         printf("\nErrorCode2       = 0x%X", WriteResult.ErrorCode2);
         printf("\nAddData1         = 0x%X", WriteResult.AddData1);
         printf("\nAddData2         = 0x%X", WriteResult.AddData2);
      }
      else
      {
         printf("\nWrite RecordData failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}

//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step10_StopProfinet(void)
{
   printf("\n\nStep 10: Stop the Profinet. Press key...");
   _getch();
   {
      // DLL function call; the function stops the Profinet
      error_class = CsPnioStopProfinet(board_nr);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nProfinet stop Controller %d OK", board_nr);
      }
      else
      {
         printf("\nProfinet Controller %d not stopped. %s", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}
