
// *******************************************************************************************************
// *******************************************************************************************************
// ************
// ************ Project:      Basic Profinet Device Example for communication with a Device - KUNBUS GmbH
// ************
// ************ Subject:      Communication with a device board by use of dfpnio.dll functions
// ************ Filename:     Basic_Profinet_Device_Example.cpp
// ************ Company:      KUNBUS GmbH, Robert-Bosch-Strasse 12a, D-76275 Ettlingen
// ************ Autor:        AME, 19.02.2015
// ************ Description:  This software is a BASIC example how to use a Profinet device board.
// ************
// ************ Expanded:     v1.0  19.02.2015 AME : initial version
// ************ Expanded:     v1.1  26.06.2015 AME : replace getchar() with _getch() and _putch()
// ************ Expanded:     v1.2  23.09.2015 CVO : Build with latest dfpnio.dll
// ************ Expanded:     v1.3  05.10.2015 CVO : Changed DLL function names
// ************ Expanded:     v1.4  14.07.2016 CVO : Change to KUNBUS, new dll
// ************
// *******************************************************************************************************
// *******************************************************************************************************
#define          VERSION_STR "v1.4"

//   What this application wants to demonstrate
//
// - Use this software parallel to the documention about the Profinet Device board.
// - This software is a BASIC example how to use a Profinet device with dfpnio.dll
// - The intention is to give a short example how the Profinet communication is working.
//   Only the absolut basic functionality is used.
// - Use the Microsoft Visual Studio Debugger to singlestep this software. Without steping it will be harder
//   to understand how the interface works.
//
//
//   Requirements for compiling and running this demo successfull on your system
//
// - Windows 7 or higher
// - Microsoft Visual Studio 2012. Other compiler will also work but are not tested. Maybe some errors could occure.
// - The setup for the Profinet controller/device driver has to be executed successfully.
// - At least 1 Profinet controller/device hardware is placed in the PC.
// - Use an Profinet controller to connecting the KUNBUS Profinet Device Board.
//
//
//   How the demo works
//
//   There are 12 steps
//   Step 1:  optional: Check the DLL version of dfpnio.dll
//   Step 2:  Get information about the number of installed boards, names of the boards and
//            information about where i nthe PC (PCI slot) the board is placed.
//   Step 3:  Choose a board number from the installed hardware.
//   Step 4:  Start Profinet.
//   Step 5:  Get information about the connected controller(s).
//   Step 6:  Read / Write Process Data (Cyclic)
//            High level function. Data on a specific Slot can be exchanged
//   Step 7:  Read / Write All Process Data (Cylcic)
//            Always the data on all Slots will be exchaged.
//            The Better Way for exchanging the process data is the example in Step 6
//   Step 8:  Read / Write Record Data (Acylcic)
//   Step 9:  Pull / Plug Modules
//            It is recommended that Slot 1 is configured by an connected controller
//   Step 10: send alarm
//   Step 11: Set / Reset Diagnosis
//   Step 12: Stop Profinet.
//
//
//   Hints and tips for programming:
// - If a new Visual Studio project is created, 1 external library dfpnio.lib has to be added to the
//   linkers library list.
//   For Visual Studio 2012: Go to project/settings/All Configurations/Linker/Object-LibraryModules/
//   there you have to add the library dfpnio.lib (separated with a blank, if other libs are already included).
// - pointer has a small "p" as the first letter in their name (for example "pControllerInfo" is a pointer,
//   "DeviceInfo" is an instance)



// **************** Standard C Header includes ***********************************************************

#ifdef __linux__
	#include <unistd.h>
	#include <string.h>
    #define _getch() getchar()
    #define _putch(a) putchar(a)
	#define Sleep(ms) usleep(ms*1000)
	#define scanf_s scanf
	#define sprintf_s sprintf
#else
	#include <Windows.h>
	#include <conio.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>


// **************** Header includes for Profinet Communication under Windows *******************************
#include "dfpnio.h"
#include "PN_LibMain.h"
#include "comm_interface_errors.h"



//*********************** Prototypes of used subfunctions ******************
void Step1_GetDLLVersion(void);
void Step2_GetBoardInfo(void);
void Step3_ChooseBoard(void);
void Step4_StartProfinet(void);
void Step5_GetControllerInformation(void);
void Step6_ReadWriteProcessData(void);
void Step7_ReadWriteAllProcessDeviceData(void);
void Step8_ReadWriteRecordData(void);
void Step9_PullPlugModule(void);
void Step10_SendAlarm(void);
void Step11_SetResetDiagnosis(void);
void Step12_StopProfinet(void);

//*********************** structures ******************
typedef struct _ioStruct
{
   INT16U DataLength;
   INT8U *pData;
   INT8U *pIoPs;
   INT8U *pIoCs;
   INT8U *pApdu;
} tIoData;


//*********************** Global variables ******************
/*
   error_class: in case of a command was not processed correctly. In case of a
   error_class != 0 additional information is found in the error value,
   which can be read by a call to GetControllerError(board_nr).
   Or full analyzed error status in text form by a call to
   GetDLLErrorText(error_class, GetControllerError(board_nr))
*/
   INT32U error_class = ERROR_CLASS_NO_ERROR;  // return value of every DLL function
   INT32U board_count = 0;                // number of Profinet controllers in the system
   INT32U board_nr    = 0;                // controller to be used in this program
   INT32U device_nr  = 0;                // device to be used in this program


//*************************************** MAIN FUNCTION **************************************************
int main( void )
{
   int Mode = 0;

   printf("\n\n\nBasic Profinet Device Example %s - KUNBUS GmbH\n\n", VERSION_STR);

   //Step 1: optional: Check DLL version
   Step1_GetDLLVersion();

   //Step 2: Get information about the number of installed boards, names and installation PCI slot
   Step2_GetBoardInfo();

   //Step 3: Choose a board
   Step3_ChooseBoard();

   do {
      printf("\n\nCommand");
      printf("\n 0 - Exit");
      printf("\n 1 - Identify connected Controller(s)");
      printf("\n 2 - Read/Write I/O Data (Cyclic)");
      printf("\n 3 - Read/Write Record Data (Acyclic)");
      printf("\n 4 - Pull/Plug Module");
      printf("\n 5 - Send Process Alarm");
      printf("\n 6 - Set/Reset Diagnosis");

      printf("\nSelect Command: ");
      scanf_s("%d", &Mode);

      switch (Mode)
      {
      case 0:
         // exit
         break;
      case 1: // AR infos
         //Step 4: Start Profinet
         Step4_StartProfinet();

         // Step 5: Get information about connected controller(s)
         Step5_GetControllerInformation();

         //Step 12: Stop Profinet
         Step12_StopProfinet();
         break;

      case 2: // process data
         //Step 4: Start Profinet
         Step4_StartProfinet();

         //Step 6: read and write process Data
         Step6_ReadWriteProcessData();

         // Another way for getting process data is shown in the following function.
         // But this method is not so comfortable like it is shown in "Step6_ReadWriteProcessData"
         // Step7_ReadWriteAllProcessDeviceData();

         //Step 12: Stop Profinet
         Step12_StopProfinet();
         break;

      case 3: // record data
         //Step 4: Start Profinet
         Step4_StartProfinet();

         //Step 8: read and write record data
         Step8_ReadWriteRecordData();

         //Step 12: Stop Profinet
         Step12_StopProfinet();
         break;

      case 4: // pull/plug module
         //Step 4: Start Profinet
         Step4_StartProfinet();

         // Step 9: // pull / plug module
         Step9_PullPlugModule();

         //Step 12: Stop Profinet
         Step12_StopProfinet();
         break;

      case 5: // send alarm
         //Step 4: Start Profinet
         Step4_StartProfinet();

         // Step 10: send alarm
         Step10_SendAlarm();

         //Step 12: Stop Profinet
         Step12_StopProfinet();
         break;

      case 6: // set reset diagnosis
         //Step 4: Start Profinet
         Step4_StartProfinet();

         // Step 10: send alarm
         Step11_SetResetDiagnosis();

         //Step 12: Stop Profinet
         Step12_StopProfinet();
         break;

      default:
         printf("\n Invalid Input \n");
         break;
      }
   }while(Mode != 0);

   printf("\n\n\nFinished Basic Profinet Example. Press key ...." );
   _getch();
   return(0);
}



//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step1_GetDLLVersion(void)
{
   printf("\n\nStep 1 (optional): Check DLL version. Press key...");
   _getch();
   {
      char* pversion_string;
      // DLL function call; the string is managed by the DLL
      CsPnioGetDllVersion(&pversion_string); // return value not used, because this DLL function is always successful
      printf("\nDLL Version used is : %s", pversion_string);
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step2_GetBoardInfo(void)
{
   printf("\n\nStep 2: Get board info. Press key...");
   _getch();
   {
      CONTROLLER_INFO* pControllerInfo;
      // DLL function call; the memory of the controller info structure is managed by the DLL
      error_class = CsPnioGetControllerInfo(&board_count, &pControllerInfo);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         INT32U i;
         for (i = 0; i < board_count; i++)
         {
            printf("\nBoard %d  named: %s  placed in PCI bus/device/function %d",
               pControllerInfo->Index, pControllerInfo->Name, pControllerInfo->PCIslot);
            pControllerInfo++;
         }
      }
      else
      {
         printf("\nGet board info failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(0)));
      }
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step3_ChooseBoard(void)
{
   printf("\n\nStep 3: Choose a board. Press key...");
   _getch();
   // the board_count is known after step 2
   if (board_count > 0)
   {
      INT32U KeybordInput;
      do
      {
         printf("\r\nEnter board number 0 to %d: ", board_count - 1);
         KeybordInput = scanf_s("%d", &board_nr);
         if (KeybordInput  == 0) //empty keyboardbuffer if value was not readable
         {
            scanf_s("%d", &KeybordInput);
            KeybordInput=0;
         }
      }while ((KeybordInput == 0) || (board_nr > board_count - 1));
      printf("Board selected: %d", board_nr);
   }
   else
   {
      printf("\nNo controller installed or driver not started");
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step4_StartProfinet(void)
{
   printf("\n\nStep 4: Start the Profinet. Press key...");
   _getch();
   {
      // DLL function call; the function initializes some DLL internal structures and starts the Profinet
      error_class = CsPnioStartProfinetDevice(board_nr);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nDevice Profinet start on Board %d OK", board_nr);
      }
      else
      {
         printf("\nDevice Profinet on Board %d not started. %s.", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step5_GetControllerInformation(void)
{
   printf("\n\nStep 5: Getting information about the connected Controller(s). Press key...");
   _getch();
   {
      INT16U NumberOfAr = 0;
      DFPN_AR_INFO *pArInfo;

      char repeatGetArInfo = 'Y';

      do
      {
         repeatGetArInfo = 'Y';
         // DLL function call; the function returns how many profinet controllers are connected to the device
         // and gives more information about the controller(s)
         error_class = CsPnioDeviceGetArInfo(board_nr,
                                       &NumberOfAr,
                                       &pArInfo);
         printf("\n%d Controller are connected to Board %d", NumberOfAr, board_nr);

         if (error_class == ERROR_CLASS_NO_ERROR)
         {
            //printf("\n%d Controller are connected to Board %d ", NumberOfAr, board_nr);
            INT16U ArCnt = 0;

            while(ArCnt < NumberOfAr)
            {
               printf("\nController %d", ArCnt+1);
               printf("\n   DNS: %s", pArInfo[ArCnt].CMInitName);
               printf("\n   IP:  %d.%d.%d.%d", pArInfo[ArCnt].CMInitIPAddr[0],
                                               pArInfo[ArCnt].CMInitIPAddr[1],
                                               pArInfo[ArCnt].CMInitIPAddr[2],
                                               pArInfo[ArCnt].CMInitIPAddr[3]);
               printf("\n   MAC: %02X:%02X:%02X:%02X:%02X:%02X",  pArInfo[ArCnt].CMInitMacAddr[0],
                                                                  pArInfo[ArCnt].CMInitMacAddr[1],
                                                                  pArInfo[ArCnt].CMInitMacAddr[2],
                                                                  pArInfo[ArCnt].CMInitMacAddr[3],
                                                                  pArInfo[ArCnt].CMInitMacAddr[4],
                                                                  pArInfo[ArCnt].CMInitMacAddr[5]);
               printf("\n   DAP: ");   /*!< TRUE if this AR has a configured DAP */
               if(pArInfo[ArCnt].DAPConfigured != '0')
               {
                  printf("Yes");   /*!< Yes if this AR has a configured DAP */
               }
               else
               {
                  printf("No");   /*!< No if this AR has no configured DAP */
               }

               ArCnt++;
            }
         }
         else
         {
            printf("\nGet Controller information on Board %d not started. %s.", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
         }

         printf("\nCheck again for connected Controller? [Y/n]: ");
         repeatGetArInfo = _getch();
         _putch(repeatGetArInfo);
      }while(repeatGetArInfo != 'n' && repeatGetArInfo != 'N');
   }
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************

void Step6_ReadWriteProcessData(void)
{
   // The status defines are from PN_LibMain.h
   const INT8U APDUstatusOK = DS_STATION_NO_PROBLEM | DS_PROV_STATE_RUN | DS_DATA_VALID | DS_STATE_PRIMARY;

   char repeatGetCyclicData = 'Y';

   INT16U SlotNumber = 1;
   const INT16U SubslotNumber = 1; // Only Sub-Slot 1 exists on KUNBUS Profinet Device
   INT32U SubmoduleType = 0;
   tIoData DeviceIn = {0};
   tIoData DeviceOut = {0};
   int i;

   // Test data
   INT8U TestVal = 0x77;
   INT32U valSize = sizeof(TestVal); // 1 Byte in this case

   printf("\n\nStep 6: Read/Write Cyclic Data. Press key...");
   _getch();

   // Setup Device-Pointer
   // get the process image: write the output values to the controller and receive the current input data from it
   // This first Call may return an error that the cyclic data are not yet initialized.
   // That is ok, because the process image maybe is not yet available.
   error_class = CsPnioRW_DeviceProcessData(board_nr);
   // if (error_class != ERROR_CLASS_NO_ERROR)
   // {
   //    printf("\nThe Process image is not yet available.");
   // }
   error_class = CsPnioRW_DeviceProcessData(board_nr); // Call twice on very first call, because the first call may initalize local structures

   // Init data
   // Calling "GetInputPointer" and "GetOutputPointer" is only necessary on start and if process image has changed
   error_class = CsPnioGetInputPointer( board_nr,
                                          INTERN_DEV_OFFSET,
                                          SlotNumber,
                                          SubslotNumber,
                                          &SubmoduleType,
                                          &DeviceIn.DataLength,
                                          &DeviceIn.pData,
                                          &DeviceIn.pIoPs,
                                          &DeviceIn.pIoCs,
                                          &DeviceIn.pApdu);
   error_class |= CsPnioGetOutputPointer(board_nr,
                                          INTERN_DEV_OFFSET,
                                          SlotNumber,
                                          SubslotNumber,
                                          &SubmoduleType,
                                          &DeviceOut.DataLength,
                                          &DeviceOut.pData,
                                          &DeviceOut.pIoPs,
                                          &DeviceOut.pIoCs,
                                          &DeviceOut.pApdu);
   if (error_class != ERROR_CLASS_NO_ERROR)
   {
      printf("\nGet Output Pointer failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      return;
   }
   if (DeviceIn.pData  == NULL || DeviceIn.pApdu  == NULL || DeviceIn.pIoPs  == NULL || DeviceIn.pIoCs  == NULL ||
       DeviceOut.pData == NULL || DeviceOut.pApdu == NULL || DeviceOut.pIoPs == NULL || DeviceOut.pIoCs == NULL)
   {
      printf(" Not all pointers for IO Device could be initialized\r\n");
      return;
   }

   // Set Input Data of the device vaild
   *DeviceIn.pApdu = APDUstatusOK;
   *DeviceIn.pIoPs = GOOD;
   *DeviceIn.pIoCs = GOOD;
   memcpy(DeviceIn.pData, (const void *)&TestVal, valSize);
   // update process image and read write new data
   error_class = CsPnioRW_DeviceProcessData(board_nr);

   do
   {
      // Read Data from local buffer
      {
         if (error_class == ERROR_CLASS_NO_ERROR) //
         {
            printf("\nAPDU status:      0x%X", *DeviceOut.pApdu);

            if (APDUstatusOK != *DeviceOut.pApdu)
            {
               printf("\n!!! BAD APDU-Status: Output data are NOT valid !!!");
            }
            else
            {
               printf("\nProvider status:  0x%X", *DeviceOut.pIoPs);
               if (*DeviceOut.pIoPs == GOOD)
               {
                  printf(" (GOOD)");
               }
               else
               {
                  printf(" (BAD)");
               }
               printf("\nConsumer status:  0x%X", *DeviceOut.pIoCs);
               if (*DeviceOut.pIoCs == GOOD)
               {
                  printf(" (GOOD)");
               }
               else
               {
                  printf(" (BAD)");
               }

               printf("\nOutput length: 0x%X", DeviceOut.DataLength);
               printf("\nOutput Data (Hex) = ");
               printf("\n                 ");
               for (i = 0; i < DeviceOut.DataLength; i++)
               {
                  printf("%02X ", DeviceOut.pData[i]);
                  if ( !((i+1) % 16 ))
                  {
                     printf("\n                 ");
                  }
               }
            }
         }
      }
      // Write Data into local buffer
      {
         TestVal ++; // increase only to have different data
         // Set the Input Data of the device, to send them to the controller later
         memcpy(DeviceIn.pData, (const void *)&TestVal, valSize);

         if (error_class == ERROR_CLASS_NO_ERROR) //
         {
            printf("\nAPDU status:      0x%X", *DeviceIn.pApdu);

            if (APDUstatusOK != *DeviceIn.pApdu)
            {
               printf("\n!!! BAD APDU-Status: Input data are NOT valid !!!");
            }
            else
            {
               printf("\nProvider status:  0x%X", *DeviceIn.pIoPs);
               if (*DeviceIn.pIoPs == GOOD)
               {
                  printf(" (GOOD)");
               }
               else
               {
                  printf(" (BAD)");
               }
               printf("\nConsumer status:  0x%X", *DeviceIn.pIoCs);
               if (*DeviceIn.pIoCs == GOOD)
               {
                  printf(" (GOOD)");
               }
               else
               {
                  printf(" (BAD)");
               }

               printf("\nInput length: 0x%X", DeviceIn.DataLength);
               printf("\nInput Data (Hex)  = ");
               printf("\n                 ");
               for (i = 0; i < DeviceIn.DataLength; i++)
               {
                  printf("%02X ", DeviceIn.pData[i]);
                  if ( !((i+1) % 16 ))
                  {
                     printf("\n                 ");
                  }
               }
            }
         }
      }

      // update process image and read write new data
      // Controller reads and writes new data from the device board actively.
      // When the new data will be get or set is handled by the controller!
      error_class = CsPnioRW_DeviceProcessData(board_nr);
      if (error_class != ERROR_CLASS_NO_ERROR)
      {
         // Check first if process image has changed!
         INT32U ErrorNr = CsPnioGetControllerError(board_nr);
         if (error_class == ERROR_CLASS_DLL && ErrorNr == RET_DLL_CYCLIC_DATA_NOT_INITIALIZED)
         {
            printf("\nProcess Image changed. Updateing pointer...");

            error_class = CsPnioGetInputPointer( board_nr,
                                                   INTERN_DEV_OFFSET,
                                                   SlotNumber,
                                                   SubslotNumber,
                                                   &SubmoduleType,
                                                   &DeviceIn.DataLength,
                                                   &DeviceIn.pData,
                                                   &DeviceIn.pIoPs,
                                                   &DeviceIn.pIoCs,
                                                   &DeviceIn.pApdu);
            error_class |= CsPnioGetOutputPointer(board_nr,
                                                   INTERN_DEV_OFFSET,
                                                   SlotNumber,
                                                   SubslotNumber,
                                                   &SubmoduleType,
                                                   &DeviceOut.DataLength,
                                                   &DeviceOut.pData,
                                                   &DeviceOut.pIoPs,
                                                   &DeviceOut.pIoCs,
                                                   &DeviceOut.pApdu);
            if (error_class != ERROR_CLASS_NO_ERROR)
            {
               printf("\nGet Output Pointer failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
               break;
            }
            if (DeviceIn.pData  == NULL || DeviceIn.pApdu  == NULL || DeviceIn.pIoPs  == NULL || DeviceIn.pIoCs  == NULL ||
                DeviceOut.pData == NULL || DeviceOut.pApdu == NULL || DeviceOut.pIoPs == NULL || DeviceOut.pIoCs == NULL)
            {
               printf(" Not all pointers for IO Device could be initialized\r\n");
               break;
            }

            // Set Input Data of the device vaild
            *DeviceIn.pApdu = APDUstatusOK;
            *DeviceIn.pIoPs = GOOD;
            *DeviceIn.pIoCs = GOOD;
            memcpy(DeviceIn.pData, &TestVal, valSize);
         }
         else
         {
            printf("\nSetup Device Pointer failed in RW_DeviceProcessData.");
            break;
         }
      }


      printf("\nRead/Write again Cyclic Data? [Y/n]: ");
      repeatGetCyclicData = _getch();
      _putch(repeatGetCyclicData);

   }while(repeatGetCyclicData != 'n' && repeatGetCyclicData != 'N');
}





//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
// works only for one Controller !!!
// The better way is shown in example "Step6_ReadWriteProcessData"
void Step7_ReadWriteAllProcessDeviceData(void)
{
   // The status defines are from PN_LibMain.h
   const INT8U APDUstatusOK = DS_STATION_NO_PROBLEM | DS_PROV_STATE_RUN | DS_DATA_VALID | DS_STATE_PRIMARY;

   char repeatGetCyclicData = 'Y';
   INT8U data_byte = 0x77;
   DEVICE_INFO DeviceInfo;
   int i;

   printf("\n\nStep 7: Read/Write All Cyclic Data. Press key...");
   _getch();

   // get the process image: write the output values to the controller and receive the current input data from it
   // This first Call may return an error that the cyclic data are not yet initialized.
   // That is ok, because the process image maybe is not yet available.
   error_class = CsPnioRW_DeviceProcessData(board_nr);
   // if (error_class != ERROR_CLASS_NO_ERROR)
   // {
   //    printf("\nThe Process image is not yet available.");
   // }

   // the first call of GetDeviceData is used to initalize DeviceInfo.
   // We will get pointer and length information to write inputs
   error_class = CsPnioGetDeviceData(board_nr,
                               INTERN_DEV_OFFSET, // If the Profinet board is used as Device, use this #define as device parameter.
                               &DeviceInfo);
   if (error_class != ERROR_CLASS_NO_ERROR)
   {
      printf("\nGetDeviceData failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      return;
   }

   // Set Input Data of the current device vaild
   DeviceInfo.InputsStatus = GOOD;  // Input Provider Status is GOOD

   do
   {
      // Set the Input Data of the device, to send them to the controller later
      memset(DeviceInfo.pInputData, data_byte, DeviceInfo.InputsLength);
      data_byte ++; // increase only to have different data

      // DLL function call; the data are written to DLL internal structures, not yet to the controller
      error_class = CsPnioSetDeviceData(board_nr, INTERN_DEV_OFFSET, &DeviceInfo);
      if (error_class != ERROR_CLASS_NO_ERROR)
      {
         printf("\nSet device data failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
         break;
      }

      // GetDeviceData is a function which works actually only for one Controller!!!
      // It provides all data of all slots of the chosen board one after the other.
      // This means that you have to know which slot has which size to work with the data.
      // If there are more than one Controller, than you don't know which controller starts on which place.
      // So you can not know which information is on which place.
      //    Another way to get or set the data is to use the function "GetInputPointer" and "GetOutputPointer"
      //    where you can get/set the data of a specific slot/subslot.
      //    In case of using the "shared device" functionality, you know from which controller your data are,
      //    because you did the configuration by yourself!
      //    I.e.: Slots 1 and 2 are Configured in Controller 1 (Maybe: KUNBUS Profinet Controller board)
      //          Slots 3 and 4 are Configured in Controller 2 (Maybe: Siemens S7)
      error_class = CsPnioGetDeviceData(board_nr,
                                 INTERN_DEV_OFFSET,
                                 &DeviceInfo);

      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nAPDU status:   0x%X", DeviceInfo.APDUStatus);
         printf("\nDevice status: 0x%X", DeviceInfo.DeviceStatus);
         printf("\nInput status:  0x%X", DeviceInfo.InputsStatus);
         if (DeviceInfo.InputsStatus == GOOD)
         {
            printf(" (GOOD)");
         }
         else
         {
            printf(" (BAD)");
         }
         printf("\nOutput status: 0x%X", DeviceInfo.OutputsStatus);
         if (DeviceInfo.OutputsStatus == GOOD)
         {
            printf(" (GOOD)");
         }
         else
         {
            printf(" (BAD)");
         }
         printf("\nInput length:  0x%X", DeviceInfo.InputsLength);
         printf("\nOutput length: 0x%X", DeviceInfo.OutputsLength);

         if(APDUstatusOK == DeviceInfo.APDUStatus)
         {
            printf("\n!!! BAD APDU-Status: Input and Output data are NOT valid !!!");
         }
         printf("\nInput Data (Hex) = ");
         printf("\n               ");
         for (i = 0; i < DeviceInfo.InputsLength; i++)
         {
            printf("%02X ", DeviceInfo.pInputData[i]);
            if ( !((i+1) % 16 ))
            {
               printf("\n               ");
            }
         }

         printf("\nOutput Data (Hex)  = ");
         printf("\n               ");
         for (i = 0; i < DeviceInfo.OutputsLength; i++)
         {
            printf("%02X ", DeviceInfo.pOutputData[i]);
            if ( !((i+1) % 16 ))
            {
               printf("\n               ");
            }
         }

      }
      else
      {
         printf("\nGet Cyclic Data failed on Board %d. %s", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
         break;
      }

      // refresh the process image: write the output values to the controller and receive the current input data
      error_class = CsPnioRW_DeviceProcessData(board_nr);
      if (error_class != ERROR_CLASS_NO_ERROR)
      {
         printf("\nProcess image access failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
         break;
      }

      printf("\nRead/Write again Cyclic Data? [Y/n]: ");
      repeatGetCyclicData = _getch();
      _putch(repeatGetCyclicData);

   }while(repeatGetCyclicData != 'n' && repeatGetCyclicData != 'N');
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step8_ReadWriteRecordData(void)
{
   char repeatAcyclicData = 'Y';

   INT8U  pRecordReadData[1440] = {0};    // We use max size, because we don't know the real size before calling DeviceGetRecordData.
   INT16U recordReadDataLen = 1440;       // actually size will be returned by function "DeviceGetRecordData"
   INT8U  pRecordSendData[] = {0x00, 0x01, 0x02, 0x03, 0xFF}; // some Hex values which will be interpreted by the controller
   INT16U recordSendDataLen = sizeof pRecordSendData;

   INT16U SlotNr = 1;
   INT8U  ReadAccess = 0x1;  // 0 = Read  access denied, 1 = Read  access granted
   INT8U  WriteAccess = 0x1; // 0 = Write access denied, 1 = Write access granted
   int i;

   printf("\n\nStep 8: Read/Write Record Data. Press key...");
   _getch();

   do
   {
      // Read Record Data from the Controller
      // Records will always be read from an internal memory.
      // The controller will set the record data to the inernal memory
      error_class = CsPnioDeviceGetRecordData(board_nr,
                                        SlotNr,
                                        0,                  // only one index allowed per Slot
                                        &ReadAccess,        // returned: which Access is granted
                                        &WriteAccess,       // returned: which Access is granted
                                        pRecordReadData,    // returned: data of record
                                        &recordReadDataLen);// returned: real length of record data

      printf("\nRead Record Data:");
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n   Read  access on Slot %d: %d ", SlotNr, ReadAccess);
         if(0x1 == ReadAccess)
         {
            printf("(granted)");
         }
         else
         {
            printf("(denied)");
         }

         printf("\n   Write access on Slot %d: %d ", SlotNr, WriteAccess);
         if(0x1 == WriteAccess)
         {
            printf("(granted)");
         }
         else
         {
            printf("(denied)");
         }

         if(0x1 == ReadAccess)
         {
            printf("\n   Read record data (Hex)  = ");
            printf("\n               ");
            for (i = 0; i < recordReadDataLen; i++)
            {
               printf("%02X ", pRecordReadData[i]);
               if ( !((i+1) % 16 ))
               {
                  printf("\n               ");
               }
            }
         }
         else
         {
            printf("\n   Read record data: Not valid - Access denied!");
         }
      }
      else
      {
         printf("\n   Read Record Data failed on Board %d. %s", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }


      ReadAccess = 0x1;  // Read  access granted
      WriteAccess = 0x1; // Write access denied

      // Write Record Data to the controller.
      // Records will always be written to an internal memory.
      // The controller will get the record data from the inernal memory
      error_class = CsPnioDeviceSetRecordData(board_nr,
                                        SlotNr,
                                        0,                  // only one index allowed per Slot
                                        ReadAccess,         // Which access has the controller to the record data
                                        WriteAccess,        // Which access has the controller to the record data
                                        pRecordSendData,    // Record data for to the controller
                                        recordSendDataLen); // Length of the current record data

      printf("\nWrite Record Data:");
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n   Read  access on Slot %d: %d ", SlotNr, ReadAccess);
         if(0x1 == ReadAccess)
         {
            printf("(granted)");
         }
         else
         {
            printf("(denied)");
         }

         printf("\n   Write access on Slot %d: %d ", SlotNr, WriteAccess);
         if(0x1 == WriteAccess)
         {
            printf("(granted)");
         }
         else
         {
            printf("(denied)");
         }

         if(0x1 == WriteAccess)
         {
            printf("\n   Write record data (Hex)  = ");
            printf("\n               ");
            for (i = 0; i < recordSendDataLen; i++)
            {
               printf("%02X ", pRecordSendData[i]);
               if ( !((i+1) % 16 ))
               {
                  printf("\n               ");
               }
            }
         }
         else
         {
            printf("\n   Write record data: Not possible - Access denied!");
         }
      }
      else
      {
         printf("\n   Read Record Data failed on Board %d. %s", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
         return;
      }

      // increasing test data for next round
      for(i=0; i < recordSendDataLen; i++)
      {
          pRecordSendData[i] ++;
      }

      printf("\nRead/Write again Acyclic Data? [Y/n]: ");
      repeatAcyclicData = _getch();
      _putch(repeatAcyclicData);

   }while(repeatAcyclicData != 'n' && repeatAcyclicData != 'N');
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step9_PullPlugModule(void)
{
   INT32U Slot_1 = 1;      // It is recommended that Slot 1 is configured by an connected controller
   const char PULL = 0x0;
   const char PLUG = 0x1;

   char repeatPullPlugModules = 'Y';

   printf("\n\nStep 9: Pull/Plug Modules. Press key...");
   _getch();

   do
   {
      // It is recommended that Slot 1 is configured by an connected controller

      printf("\nPull module %d. Press key...", Slot_1);
      _getch();
      // Pull configured (by controller) module 1
      // Alarm will be send to the controller...
      error_class = CsPnioDeviceChangeModState(board_nr,
                                         Slot_1,
                                         PULL);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
          printf("\n     successful.");
      }
      else
      {
         printf("\n     failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }


      printf("\nPlug module %d. Press key...", Slot_1);
      _getch();
      // Plug configured (by controller) module 1
      error_class = CsPnioDeviceChangeModState(board_nr,
                                         Slot_1,
                                         PLUG);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
          printf("\n     successful.");
      }
      else
      {
         printf("\n     failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }


      printf("\nPull/Plug Modules again? [Y/n]: ");
      repeatPullPlugModules = _getch();
      _putch(repeatPullPlugModules);

   }while(repeatPullPlugModules != 'n' && repeatPullPlugModules != 'N');
}



//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step10_SendAlarm(void)
{
   char repeatSendAlarm = 'Y';
   INT16U Slot = 1;
   INT16U AlarmType = DFPN_ALARM_TYPE_PROCESS;
   INT8U * pAlarmItem = 0;
   INT16U AlarmItemLength = 0;

   printf("\n\nStep 10: Send an alarm. Press key...");
   _getch();

   do
   {
      printf("\nSending Alarm ...");
      error_class = CsPnioDeviceSendAlarm(board_nr,
                                    Slot,
                                    AlarmType,
                                    pAlarmItem,
                                    AlarmItemLength);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n        successful.");
      }
      else
      {
         printf("\n        failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }

      printf("\nSend Alarm again? [Y/n]: ");
      repeatSendAlarm = _getch();
      _putch(repeatSendAlarm);

   }while(repeatSendAlarm != 'n' && repeatSendAlarm != 'N');
}


//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step11_SetResetDiagnosis(void)
{
   char repeatSetResetDiagnosis = 'Y';
   INT16U Slot = 1;
   DFPN_DIAG_INFO DiagItem;

   printf("\n\nStep 11: Set/Reset a Diagnosis. Press key...");
   _getch();

   memset(&DiagItem, 0, sizeof(DiagItem));
   DiagItem.UserStructureIdentifier = DFPN_DIAG_CHAN_DIAG;
   DiagItem.ChannelErrorType = DFPN_ALARM_CET_OVERTEMPERATURE;
   DiagItem.RTATrigger = DFPN_ALARM_LOW_RTA_TRIGGER;
   DiagItem.ChannelProperties = DFPN_ALARM_CP_SPECIFIER_APPEARS;

   do
   {
      printf("\nSet Diagnosis. Press key...");
      _getch();
      error_class = CsPnioDeviceSetDiag(  board_nr,
                                    Slot,
                                    &DiagItem);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n    successful.");
      }
      else
      {
         printf("\n    failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }

      printf("\nReset Diagnosis. Press key...");
      DiagItem.ChannelProperties = DFPN_ALARM_CP_SPECIFIER_DISAPPEARS;
      _getch();
      error_class = CsPnioDeviceSetDiag(  board_nr,
                                    Slot,
                                    &DiagItem);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n    successful.");
      }
      else
      {
         printf("\n    failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }

      DiagItem.ChannelProperties = DFPN_ALARM_CP_SPECIFIER_APPEARS;
      printf("\nSet Diagnosis. Press key..."); //reset Diagnosis for a special slot
      _getch();
      error_class = CsPnioDeviceSetDiag(  board_nr,
                                    Slot,
                                    &DiagItem);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n    successful.");
      }
      else
      {
         printf("\n    failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }


      printf("\nReset all Diagnosis. Press key...");
      _getch();
      error_class = CsPnioDeviceDiagReset(board_nr); // Reset all diagnosis for the complete device
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\n    successful.");
      }
      else
      {
         printf("\n    failed. %s", CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }

      printf("\nSet/Reset a Diagnosis again? [Y/n]: ");
      repeatSetResetDiagnosis = _getch();
      _putch(repeatSetResetDiagnosis);

   }while(repeatSetResetDiagnosis != 'n' && repeatSetResetDiagnosis != 'N');
}



//***********************************************************************************************************************
//***********************************************************************************************************************
//***********************************************************************************************************************
void Step12_StopProfinet(void)
{
   printf("\n\nStep 12: Stop the Profinet. Press key...");
   _getch();
   {
      // DLL function call; the function stops the Profinet
      error_class = CsPnioStopProfinetDevice(board_nr);
      if (error_class == ERROR_CLASS_NO_ERROR)
      {
         printf("\nDevice Profinet stop on Board %d OK", board_nr);
      }
      else
      {
         printf("\nDevice Profinet on Board %d not stopped. %s", board_nr, CsPnioGetDLLErrorText(error_class, CsPnioGetControllerError(board_nr)));
      }
   }
}

