/*!
 *
 * Project: Bus Firmware
 * (C)    : KUNBUS GmbH, Robert-Bosch-Strasse 12a, D-76275 Ettlingen
 *
 *
 * \file comm_interface.h Profinet
 *
 * \brief Contains communication structures which are necessary to be known in the interface.
 *
 */

#ifndef COMM_INTERFACE_H
#define COMM_INTERFACE_H

#pragma pack(push, Kunbus, 4)

#include "PN_LibMain.h"

/*****************************************************************************
 ****  driver specific stuff                                              ****
 *****************************************************************************/

/*****************************************************************************
 ****  DEFINES                                                            ****
 *****************************************************************************/


/****  DEFINES AFFECTING MEMORY AND PERFORMANCE                         ****/
/*** INBOUND Window **/ // Memory on the PCI-Card
#define DEFAULT_INBOUND_WINDOW_BASE      0x38000000
#define OFFSET_INBOUND_INTERFACE         0x00000000                    /*  0 MB */
#define OFFSET_INBOUND_INTERFACE_DEV     0x00100000                    /*  1 MB */
#define OFFSET_INBOUND_CYCLIC_DATA       0x00800000                    /*  8 MB */
#define OFFSET_INBOUND_CYCLIC_DATA_DEV   0x00820000                    /*  8 MB + 128 kB */
#define OFFSET_INBOUND_GLOBALS           0x00900000                    /*  9 MB */
#define MAX_SIZE_OF_INBOUND_WINDOW       0x01000000                    /* 16 MB */

/*** OUTBOUND Window **/ // Memory on PC (RAM)
#define DEFAULT_OUTBOUND_WINDOW_BASE     0x80000000
#define OFFSET_OUTBOUND_CYCLIC_DATA      0x00000000                    /*  0 MB */
#define OFFSET_OUTBOUND_CYCLIC_DATA_DEV  0x00020000                    /*  0 MB + 128 kB */
#define OFFSET_OUTBOUND_GLOBALS          0x000F0000                    /*  1 MB - 64k */
#define OFFSET_OUTBOUND_GLOBALS_DEV      0x000F8000                    /*  1 MB - 32k */
#define MAX_SIZE_OF_OUTBOUND_WINDOW      0x00100000                    /*  1 MB */

#define MAX_SIZE_OF_INTERFACE_CHANNEL    0x00100000                    /*  1 MB */
#define MAX_SIZE_OF_PROCESSIMAGE         0x00004000                    /* 16 k  */  // max memory for exchange data
#define MAX_SIZE_OF_DEVPROCESSIMAGE      0x00004000                    /* 16 k  */  // max memory for exchange data
#define MAX_SIZE_OF_DEVPROCESSIMAGE_HEADER  (256+32)                   /* Header with status bytes */
#define MAX_DEVICES_COUNT                128
#define MAX_IDENTIFY_ALL_DEVICES_COUNT   256
#define HIGHEST_SLOT_NUM                 128
#define MAX_NUM_OF_SLOT                  (HIGHEST_SLOT_NUM+1)        // Number of available Slots (including DAP)
#define MAX_NUM_OF_SUBMODS               (MAX_NUM_OF_SLOT+2)         // Number of allowed plugged modules (DAP has 3, all other 1)
#define ALARM_TIMER_TIMEOUT              5                           // soconds Timeout for Alarm ack
#define MAX_DEVICE_RECORD_SIZE           1440                        // Length of RW Records
#define MAX_NUMBER_RECORDS_PER_SLOT      1                           // Number of records per slot
#define DEVICE_RECORD_INDEX              0x1000                      // First Record Index
#define MAX_NUM_OF_AR                    20                          // Maximum number of controller for shared device
#define DMA_ALIGNED_BUF_SIZE             (2*3*2*32)                  // 32 byte aligned buffer for dma

#define MAX_DEVICE_INSTANCES 1  // the KW stack V3.5... only supports 1 device instance
                                // the KW stack V3.8... would support 8 device instances

/*****************************************************************************
 ****  TYPEDEFS                                                           ****
 *****************************************************************************/


/**
 * @defgroup commifpn Communication Interface for Profinet
 * Communication structures for Profinet modules
 * @{
 */


typedef struct TAG_INTERFACE_STD_HEADER_TYPE // Header for comm between usaerApp and PCI-Card
{
   volatile INT32U Command;                     // 0, Command's defined as CMD_... in this file
   volatile INT32U SubCommand;                  // 4, Subcommand's defined as SUBCMD_... in this file, corresponding to CMD_... defines
   volatile INT32U InterfaceNo;                 // 8, Selectable Ethernet-Interface
   volatile INT32U Com_Type;                    //12, Communication type defined a EVENT_... in this file
   volatile INT32U ID;                          //16, BoardNo (PCI-SLOT) or IP-Address
   volatile INT32U ErrorClass;                  //20, Return Value
   volatile INT32U Error;                       //24, Error in the Communication stack (Profinet ec.)
   volatile INT32U Offset;                      //28, Start-Offset inside the Data (not used right now)
   volatile INT32U Length;                      //32, Length of the corresponding Data
   volatile INT32U Interrupt;                   //36, shared message signalled interrupt with additional information in the upper word
   volatile INT32U Reserved[2];                 //40, 16 byte alignement for 64 bit systems // Padding to have a aligned memory
} INTERFACE_STD_HEADER_TYPE, *P_INTERFACE_STD_HEADER_TYPE;
#define INTERFACE_STD_HEADER_TYPE_SIZE sizeof(INTERFACE_STD_HEADER_TYPE)

typedef struct TAG_INTERFACE_STD_HEADER_SHORT_TYPE // Header for comm between usaerApp and SPI
{
    volatile INT8U Command;               // 0, Command's defined as CMD_... in this file
    volatile INT8U SubCommand;            // 1, Subcommand's defined as SUBCMD_... in this file, corresponding to CMD_... defines
    volatile INT16U Reserved;             // 2
    volatile INT32U ErrorClass;       // 4, Return Value
    volatile INT32U Error;            // 8, Error in the Communication stack (Profinet ec.) (optional)
    volatile INT32U Length;           // 12, Length of the corresponding Data
} INTERFACE_STD_HEADER_SHORT_TYPE, *P_INTERFACE_STD_HEADER_SHORT_TYPE;
#define INTERFACE_STD_HEADER_SHORT_TYPE_SIZE sizeof(INTERFACE_STD_HEADER_SHORT_TYPE)


// interrupt format: 0x00000000
//                     |    |L message value (0 - 31)
//                     |    L_ shared interrupt value (0 - 7)
//                     L______ additional information (0 - 65535)
#define INTERRUPT_CHANNEL_CONTROLLER      0x00000000
#define INTERRUPT_CHANNEL_DEVICE          0x00000108 // irq 1 with dummy msg 8
#define INTERRUPT_CHANNEL_MASK            0x00000700 // MSI 0-7 are routed to IRQ_OUT (8-31 unused) 
#define INTERRUPT_CHANNEL_MASK_MSG        0x000000FF // MSG bit is detailed optional mesage information for each of the 8 IRQ values
#define INTERRUPT_CHANNEL_MASK_INFO       0xFFFFF000 // optional additional interrupt information for each of the 8 IRQ values


// The name "INTERFACE_TYPE" is already used in windows driver kernel world. therefore this type
// has to have a different name than INTERFACE_TYPE!
typedef struct TAG_COMMON_INTERFACE_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   volatile INT8U Data[MAX_SIZE_OF_INTERFACE_CHANNEL - INTERFACE_STD_HEADER_TYPE_SIZE];  // 1MB-(48 Byte) reserved
} COMMON_INTERFACE_TYPE, *P_COMMON_INTERFACE_TYPE;
#define COMMON_INTERFACE_TYPE_SIZE sizeof(COMMON_INTERFACE_TYPE)


typedef struct TAG_PNIO_PROCESSIMAGE_TYPE
{
   volatile INT8U ApduStatus[MAX_DEVICES_COUNT];          //The APDU status = DataStatus is part of the profinet telegramm on the ethernetbus.
                                                                  //The following bits are defined in the PN_LibMain.h
                                                                  //#define DS_STATE_BACKUP                     0x0000
                                                                  //#define DS_STATE_PRIMARY                    0x0001
                                                                  //#define DS_DATA_INVALID                     0x0000
                                                                  //#define DS_DATA_VALID                       0x0004
                                                                  //#define DS_PROV_STATE_STOP                  0x0000
                                                                  //#define DS_PROV_STATE_RUN                   0x0010
                                                                  //#define DS_STATION_PROBLEM                  0x0000
                                                                  //#define DS_STATION_NO_PROBLEM               0x0020     Wireshark: "StationProblemIndicator (1:Ok/0:Problem)"
   volatile INT8U DeviceStatus[MAX_DEVICES_COUNT];        //The Device status is constructed in the Kunbus Profinet Controller. The first two bits are identical
                                                                  //to the Alarm class bits
                                                                  #define DIAGNOSIS_ALARM                       0x01   //SUBCMD_ALARM_TYPE
                                                                  #define PROCESS_ALARM                         0x02   //SUBCMD_ALARM_TYPE
                                                                  #define MODULE_DIFF_BLOCK                     0x04   //SUBCMD_DIAG_MODULE_DIFF_BLOCK_TYPE
                                                                  #define UDP_DATA_AVAILABLE                    0x08   //CMD_UDP_USER_TYPE This flag is used in device Controller DeviceStatus[0]
                                                                  //        unused                             0xFC
   volatile INT8U ProcessImage[MAX_SIZE_OF_PROCESSIMAGE - MAX_DEVICES_COUNT - MAX_DEVICES_COUNT]; //Processdata including Provider an consumer status for subslots.
} PNIO_PROCESSIMAGE_TYPE, *P_PNIO_PROCESSIMAGE_TYPE;
#define PNIO_PROCESSIMAGE_TYPE_SIZE sizeof(PNIO_PROCESSIMAGE_TYPE)

// todo: in PNIO_user.h auslagern
// this type is used for application communication with the driver
// the driver uses the header to return an Error and otherwise
// copies the data by DMA into the FW simpson buffers.
typedef struct TAG_CYCLIC_PROCESSIMAGE_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   PNIO_PROCESSIMAGE_TYPE    CyclicData;
} CYCLIC_PROCESSIMAGE_TYPE, *P_CYCLIC_PROCESSIMAGE_TYPE;
#define CYCLIC_PROCESSIMAGE_TYPE_SIZE sizeof(CYCLIC_PROCESSIMAGE_TYPE)

typedef struct TAG_INBOUND_CYCLIC_DATA_TYPE // in PCI card
{
   volatile INT8U outp_latest;
   volatile INT8U outp_reading;
   volatile INT8U outp_slot[2];
   PNIO_PROCESSIMAGE_TYPE OutputSimpsonData[2][2]; // writing is done by PCI-Controler (userApp) and ProfiNet-Controler is reading  - from Master to slave / Controler to Device
   PNIO_PROCESSIMAGE_TYPE InputData;
   volatile INT8U inp_latest;
   volatile INT8U inp_reading;
   volatile INT8U inp_slot[2];
} INBOUND_CYCLIC_DATA_TYPE, *P_INBOUND_CYCLIC_DATA_TYPE;
#define INBOUND_CYCLIC_DATA_TYPE_SIZE sizeof(INBOUND_CYCLIC_DATA_TYPE)

typedef struct TAG_OUTBOUND_CYCLIC_DATA_TYPE
{
   volatile INT8U inp_latest;
   volatile INT8U inp_reading;
   volatile INT8U inp_slot[2];
   PNIO_PROCESSIMAGE_TYPE InputSimpsonData[2][2]; // writing is done by ProfiNet-Controler and PCI-Controler (UserApp) is reading  - from Slave to Master / Device to Controler
   PNIO_PROCESSIMAGE_TYPE OutputData;
   volatile INT8U outp_latest;
   volatile INT8U outp_reading;
   volatile INT8U outp_slot[2];
} OUTBOUND_CYCLIC_DATA_TYPE, *P_OUTBOUND_CYCLIC_DATA_TYPE;
#define OUTBOUND_CYCLIC_DATA_TYPE_SIZE sizeof(OUTBOUND_CYCLIC_DATA_TYPE)

typedef struct TAG_GLOBALS_DATA_TYPE // Verwaltungsdaten
{
   //All values are used in little endian format. If the card want to use them the values has to be swapped.
   //where it is possible, defines which are endinaess-save might be used.
   //The order of this values should be not changed! New values shall be added to the end to prevent incompatibilitys.
   volatile INT32U InterruptToPC_Enable;                       // Initialized by the Driver win or linux
                                                               // Used with Windows. The Windows Driver has to write Magic Number to acitvate
                                                               // Interrupt Transmission.
                                                               // Offset: 0x0
         #define INTERRUPT_TO_PC_ENABLED     0x41434341        // Ascii "ACCA"; The value looks in big and little endian identical.
   volatile INT32U ManualIntToPCTrigger;                       // Initialized by Card. Used to indicate an Interrupt to Labview. LAbview does poll this value.
                                                               // Offset: 0x4
         #define INT_TRIGGER                 0x494E4E49        // Ascii "INNI"; The value looks in big and little endian identical.
   volatile INT32U InputDataProcessImageLen;                   // Initialized by Card. Little Endian format.
                                                               // Offset: 0x8
   volatile INT32U OutputDataProcessImageLen;                  // Initialized by Card. Little Endian format.
                                                               // Offset: 0xC
   volatile INT32U PNBusRunning;                               // Initialized by Card. Little Endian format.
                                                               // Offset: 0x10
         //#define PN_BUS_ACTIV                0x11111111      // value used to read the value "UINT_32 PNBusRunning" from the c_globals
   volatile INT32U DriverType;                                 // Initialized by the Driver.
                                                               // Offset: 0x14
         #define WINDOWS_DRIVER              0x2057494E        // Ascii " WIN"
         #define LABVIEW_DRIVER              0x4C414256        // Ascii "LABV"
         #define LINUX_DRIVER                0x4C494E55        // Ascii "LINU"
         #define TCP_DRIVER                  0X20544350        // Ascii " TCP"
   volatile INT32U DriverVersion1;                             // Initialized by the Driver. HIGHEST Version
                                                               // Offset: 0x18
         #define MINIMUM_WIN_VERSION_1_PN_CONTROLLER        2
         #define MINIMUM_LV_VERSION_1_PN_CONTROLLER         0
         #define MINIMUM_LINUX_VERSION_1_PN_CONTROLLER      0
   volatile INT32U DriverVersion2;                             // Initialized by the Driver.
                                                               // Offset: 0x1C
         #define MINIMUM_WIN_VERSION_2_PN_CONTROLLER        1     
         #define MINIMUM_LV_VERSION_2_PN_CONTROLLER         0     
         #define MINIMUM_LINUX_VERSION_2_PN_CONTROLLER      0     
   volatile INT32U DriverVersion3;                             // Initialized by the Driver.
                                                               // Offset: 0x20
         #define MINIMUM_WIN_VERSION_3_PN_CONTROLLER        7     
         #define MINIMUM_LV_VERSION_3_PN_CONTROLLER         0     
         #define MINIMUM_LINUX_VERSION_3_PN_CONTROLLER      0     
   volatile INT32U DriverVersion4;                             // Initialized by the Driver. Lowest Version
                                                               // Offset: 0x24
         #define MINIMUM_WIN_VERSION_4_PN_CONTROLLER        4     
         #define MINIMUM_LV_VERSION_4_PN_CONTROLLER         0     
         #define MINIMUM_LINUX_VERSION_4_PN_CONTROLLER      0     
   volatile INT32U FW_Ready;                                   // FW ready for Commands (Dispatcher ready)
                                                               // Offset: 0x28
         #define FW_READY_INDICATION         0x11111111           
   volatile INT32U Outbound_Ready;                             // Outbound ready on PC
                                                               // Offset: 0x2C
         #define OUTBOUND_READY_INDICATION 0x11111111             
   volatile INT32U AcyclicServicePending;                      // Acyclic Service Pending
         #define ACYCLIC_SERVICE_PENDING 0x11111111            // Offset: 0x30

   //ThreadWD_xxxTask  // WD=WatchDog  // every thread inreases these variables
   volatile INT32U NotUsed0;                                   // Thread state control helper variable
   volatile INT32U NotUsed1;                                   // Thread state control helper variable
   volatile INT32U NotUsed2;                                   // Thread state control helper variable
   volatile INT32U NotUsed3;                                   // Thread state control helper variable
   volatile INT32U NotUsed4;                                   // Thread state control helper variable

   volatile INT32U Alive;                                      // This Alive value is used to monitor the card. The value is incremented
                                                               // every ms hich prio in the timer tick routine. The win and LV driver use this
                                                               // value to detect hardware failure or missing hardware. This does not work during reboot!

   volatile INT32U NotUsed5;                                   // Thread state control helper variable
   volatile INT32U PNDeviceRunning;                            // Initialized by Card. Little Endian format.
   volatile INT32U InputDataDevProcessImageLen;                // Initialized by Card. Little Endian format.
   volatile INT32U OutputDataDevProcessImageLen;               // Initialized by Card. Little Endian format.
   volatile INT32U AppAlive;                                   // The AppAlive value is used to monitor the application. It is incremented by the driver
                                                               // on processimage access
   volatile INT32U ManualIntToPCTriggerDev;                    // Initialized by Card. Used to indicate an Interrupt to Labview. LAbview does poll this value.
   volatile INT32U AppAliveCtl;                                // The AppAlive value is used to monitor the application. It is incremented by the driver
                                                               // on processimage access
   volatile INT32U HasDmaChain;                                //  is set if the firmware has a dma chain for driver
} GLOBALS_DATA_TYPE, *P_GLOBALS_DATA_TYPE;
#define GLOBALS_DATA_TYPE_SIZE sizeof(GLOBALS_DATA_TYPE)

typedef struct TAG_OUTBOUND_GLOBALS_DATA_TYPE
{
   volatile INT32U ErrorClass;                  //Return Value
   volatile INT32U Error;                       //Error in the Communication stack (Profinet ec.)
   volatile INT32U LV_SharedMemoryOffset;       //LabVIEW SharedMemory offset -> For deallocating memory with LabVIEW
} OUTBOUND_GLOBALS_DATA_TYPE, *P_OUTBOUND_GLOBALS_DATA_TYPE;
#define OUTBOUND_GLOBALS_DATA_TYPE_SIZE sizeof(OUTBOUND_GLOBALS_DATA_TYPE)


typedef struct TAG_OUTBOUND_GLOBALS_DEV_DATA_TYPE
{
   volatile INT32U ErrorClass;                  //Return Value
   volatile INT32U Error;                       //Error in the Communication stack (Profinet ec.)
} OUTBOUND_GLOBALS_DEV_DATA_TYPE, *P_OUTBOUND_GLOBALS_DEV_DATA_TYPE;
#define OUTBOUND_GLOBALS_DEV_DATA_TYPE_SIZE sizeof(OUTBOUND_GLOBALS_DEV_DATA_TYPE)


typedef struct TAG_PNIO_PROCESSIMAGE_DEV_TYPE
{
   volatile INT8U ApduStatus[MAX_DEVICES_COUNT];
   volatile INT8U DeviceStatus[MAX_DEVICES_COUNT];
      #define DEVSTATUS_MODULE_DIFFBLOCK        0x01
      #define DEVSTATUS_DIAG_PENDING            0x04
   volatile INT16U ProcessImageCounter;  // is incremented after each reconfiguration (process image mapping maybe changed)
   volatile INT8U Reserved[MAX_SIZE_OF_DEVPROCESSIMAGE_HEADER-MAX_DEVICES_COUNT-MAX_DEVICES_COUNT-2];
   volatile INT8U ProcessImage[MAX_SIZE_OF_DEVPROCESSIMAGE - MAX_SIZE_OF_DEVPROCESSIMAGE_HEADER]; //Processdata including Provider an consumer status for subslots.
} PNIO_PROCESSIMAGE_DEV_TYPE, *P_PNIO_PROCESSIMAGE_DEV_TYPE;
#define PNIO_PROCESSIMAGE_DEV_TYPE_SIZE sizeof(PNIO_PROCESSIMAGE_DEV_TYPE)

typedef struct TAG_CYCLIC_PROCESSIMAGE_DEV_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   PNIO_PROCESSIMAGE_DEV_TYPE    CyclicData;
} CYCLIC_PROCESSIMAGE_DEV_TYPE, *P_CYCLIC_PROCESSIMAGE_DEV_TYPE;
#define CYCLIC_PROCESSIMAGE_DEV_TYPE_SIZE sizeof(CYCLIC_PROCESSIMAGE_DEV_TYPE)

typedef struct TAG_INBOUND_DEV_CYCLIC_DATA_TYPE // in PCI card
{
   volatile INT8U inp_latest;
   volatile INT8U inp_reading;
   volatile INT8U inp_slot[2];
   PNIO_PROCESSIMAGE_DEV_TYPE InputSimpsonData[2][2]; // writing is done by PCI-Controller (userApp) and ProfiNet-Device is reading  - from Master to slave / Controler to Device
   PNIO_PROCESSIMAGE_DEV_TYPE OutputData;
   volatile INT8U outp_latest;
   volatile INT8U outp_reading;
   volatile INT8U outp_slot[2];
} INBOUND_DEV_CYCLIC_DATA_TYPE, *P_INBOUND_DEV_CYCLIC_DATA_TYPE;
#define INBOUND_DEV_CYCLIC_DATA_TYPE_SIZE sizeof(INBOUND_DEV_CYCLIC_DATA_TYPE)

typedef struct TAG_OUTBOUND_DEV_CYCLIC_DATA_TYPE
{
   volatile INT8U outp_latest;
   volatile INT8U outp_reading;
   volatile INT8U outp_slot[2];
   PNIO_PROCESSIMAGE_DEV_TYPE OutputSimpsonData[2][2]; // writing is done by ProfiNet-Device and PCI-Controler (UserApp) is reading  - from Slave to Master / Device to Controler
   PNIO_PROCESSIMAGE_DEV_TYPE InputData;
   volatile INT8U inp_latest;
   volatile INT8U inp_reading;
   volatile INT8U inp_slot[2];
} OUTBOUND_DEV_CYCLIC_DATA_TYPE, *P_OUTBOUND_DEV_CYCLIC_DATA_TYPE;
#define OUTBOUND_DEV_CYCLIC_DATA_TYPE_SIZE sizeof(OUTBOUND_DEV_CYCLIC_DATA_TYPE)

// This type is only used as pointer to an already existing memory area, allocated in
// DRAM of the hardware component.
typedef struct TAG_P_INBOUND_WINDOW_TYPE
{
   COMMON_INTERFACE_TYPE Interface;
   //volatile INT8U padding_1[OFFSET_INBOUND_INTERFACE_DEV - COMMON_INTERFACE_TYPE_SIZE]; // is 0
   COMMON_INTERFACE_TYPE InterfaceDev;
   volatile INT8U padding_2[OFFSET_INBOUND_CYCLIC_DATA - OFFSET_INBOUND_INTERFACE_DEV - COMMON_INTERFACE_TYPE_SIZE];
   INBOUND_CYCLIC_DATA_TYPE CyclicData;
   volatile INT8U padding_3[OFFSET_INBOUND_CYCLIC_DATA_DEV - OFFSET_INBOUND_CYCLIC_DATA - INBOUND_CYCLIC_DATA_TYPE_SIZE];
   INBOUND_DEV_CYCLIC_DATA_TYPE CyclicDataDev;
   volatile INT8U padding_4[OFFSET_INBOUND_GLOBALS - OFFSET_INBOUND_CYCLIC_DATA_DEV - INBOUND_DEV_CYCLIC_DATA_TYPE_SIZE];
   GLOBALS_DATA_TYPE Globals;
   volatile INT8U padding_5[MAX_SIZE_OF_INBOUND_WINDOW - OFFSET_INBOUND_GLOBALS - GLOBALS_DATA_TYPE_SIZE - DMA_ALIGNED_BUF_SIZE];
   volatile INT8U DmaAlignedBuf[DMA_ALIGNED_BUF_SIZE];
}INBOUND_WINDOW_TYPE, *P_INBOUND_WINDOW_TYPE;
#define INBOUND_WINDOW_TYPE_SIZE sizeof(struct TAG_P_INBOUND_WINDOW_TYPE)

// This type is only used as pointer to an already existing memory area, allocated in
// contiguous PC memory.
typedef struct TAG_P_OUTBOUND_WINDOW_TYPE
{
   OUTBOUND_CYCLIC_DATA_TYPE CyclicData;
   volatile INT8U padding_1[OFFSET_OUTBOUND_CYCLIC_DATA_DEV - OUTBOUND_CYCLIC_DATA_TYPE_SIZE];
   OUTBOUND_DEV_CYCLIC_DATA_TYPE CyclicDataDev;
   volatile INT8U padding_2[OFFSET_OUTBOUND_GLOBALS - OFFSET_OUTBOUND_CYCLIC_DATA_DEV - OUTBOUND_DEV_CYCLIC_DATA_TYPE_SIZE];
   OUTBOUND_GLOBALS_DATA_TYPE Globals;
   volatile INT8U padding_3[OFFSET_OUTBOUND_GLOBALS_DEV - OFFSET_OUTBOUND_GLOBALS - OUTBOUND_GLOBALS_DEV_DATA_TYPE_SIZE];
   OUTBOUND_GLOBALS_DEV_DATA_TYPE GlobalsDev;
   volatile INT8U padding_4[MAX_SIZE_OF_OUTBOUND_WINDOW - OFFSET_OUTBOUND_GLOBALS_DEV - OUTBOUND_GLOBALS_DATA_TYPE_SIZE];
}OUTBOUND_WINDOW_TYPE, *P_OUTBOUND_WINDOW_TYPE;
#define OUTBOUND_WINDOW_TYPE_SIZE sizeof(struct TAG_P_OUTBOUND_WINDOW_TYPE)



/*****************************************************************************
 ****  End of driver specific stuff                                       ****
 *****************************************************************************/



/*****************************************************************************
 ****  DEFINES                                                            ****
 *****************************************************************************/




/*****************************************************************************
 **** DEFINES Commands/Subcommands                                        ****
 *****************************************************************************/
#define CMD_RESET                                  0x1
#define SUBCMD_RESET_SW                            0x1
#define SUBCMD_RESET_HW                            0xff

#define CMD_PROFINET                               0x2
   #define SUBCMD_PROFINET_START                      0x1
   #define SUBCMD_PROFINET_STOP                       0x2
   #define SUBCMD_PROFINET_DEVICE_START               0x3
   #define SUBCMD_PROFINET_DEVICE_STOP                0x4
#define CMD_PN_DCP_SEND                            0x3
   #define SUBCMD_PN_DCP_SEND_IDENTIFY_ALL            0x1
   #define SUBCMD_PN_DCP_SEND_SET_NAME                0x2
   #define SUBCMD_PN_DCP_SEND_SET_IP                  0x3
   #define SUBCMD_PN_DCP_SEND_IDENTIFY_DEVICE_PROP    0x4
   #define SUBCMD_PN_DCP_SEND_CONTROL                 0x5

#define CMD_ALARM                                  0x4
   #define SUBCMD_ALARM_GET                           0x1
   #define SUBCMD_ALARM_ACK                           0x2
   #define SUBCMD_ALARM_MODE_SET                      0x3            //currently unused. Will be used to swith between autmatic Alarm acknowledge, and manual acknowledge

#define CMD_PN_RW_RECORD                           0x5
   #define SUBCMD_PN_RW_RECORD_READ                   0x1
   #define SUBCMD_PN_RW_RECORD_WRITE                  0x2

//This Command / Subcommands are handled internally by the dfpnio.dll The Controller card has no knowledge about them.
//They might be implemented in future for a TCP IP interface.
#define CMD_PN_CYCLIC_DATA_TRANSFER                0x6
   #define SUBCMD_PN_SET_GET_DATA                     0x1//   // cyclic complete process image data exchange
   #define SUBCMD_PN_GET_INPUT_DATA                   0x2//   // cyclic complete input data receipt
   #define SUBCMD_PN_SET_OUTPUT_DATA                  0x3//   // cyclic complete output data sending
   #define SUBCMD_PN_GET_OUTPUT_DATA                  0x4//   // for access to parts of the process image output data
   #define SUBCMD_PN_GET_IOCS_STATUS                  0x5//   // for access to parts of the process image input C status
   #define SUBCMD_PN_GET_IOPS_STATUS                  0x6//   // for access to parts of the process image input P status
   #define SUBCMD_PN_SET_GET_DEVICE_DATA              0x7

#define CMD_OFFSET_TABLE                           0x7
   #define SUBCMD_GET_OFFSET_TABLE_VAL_COUNT          0x1
   #define SUBCMD_GET_OFFSET_TABLE                    0x2
   #define SUBCMD_GET_PROCESSIMAGELEN                 0x3
   #define SUBCMD_GET_REMAP_TABLE                     0x4
   #define SUBCMD_ENABLE_REMAPPING                    0x5

#define CMD_HW_INFO                                0x8
   #define SUBCMD_HW_INFO_GET                         0x1
   #define SUBCMD_HW_INFO_SET                         0x2
   #define SUBCMD_HW_INFO_SET_DEBUG_LEVEL             0x3
   #define SUBCMD_HW_INFO_SET_TIME                    0x4
   #define SUBCMD_HW_INFO_SET_STATIONNAME             0x5

#define CMD_FILE_TRANSFER                          0x9
   #define SUBCMD_FLASH_SET                           0x1
   #define SUBCMD_FLASH_GET                           0x2
   #define SUBCMD_FILE_SET                            0x3
   #define SUBCMD_FILE_GET                            0x4
   #define SUBCMD_FACTORY_RESET_FLASH_DRIVE           0x5
   #define SUBCMD_FW_UPGRADE                          0x6

#define CMD_VISA_DLL                               0xA
   #define SUBCMD_VISA_DLL_SEARCH_DEVICES             0x1
   #define SUBCMD_VISA_DLL_GET_ALARM_TEXT             0x2
   #define SUBCMD_VISA_DLL_GET_ERROR_TEXT             0x3
   #define SUBCMD_VISA_DLL_GET_MODULEDIFFBLOCK_TEXT   0x4

#define CMD_DIAG                                   0xB
   #define SUBCMD_DIAG_DEVICE                         0x1
   #define SUBCMD_DIAG_MODULE_DIFF_BLOCK              0x2

#define CMD_CONTROLLER                             0xC
   #define SUBCMD_CONTROLLER_SET_PROVIDER_STATUS      0x1
   #define SUBCMD_SET_WATCHDOG                        0x2
   #define SUBCMD_GET_ARINFO                          0x3

#define CMD_DEVICE                                 0xD
   #define SUBCMD_DEVICE_SET_PROVIDER_STATUS          0x1
   #define SUBCMD_DEVICE_DIAG                         0x2
   #define SUBCMD_DEVICE_REMOVE_ALL_DIAG              0x3
   #define SUBCMD_DEVICE_SEND_ALARM                   0x4
   #define SUBCMD_DEVICE_CHANGE_MODSTATE              0x5
   #define SUBCMD_DEVICE_RECORD_SET                   0x6
   #define SUBCMD_DEVICE_RECORD_GET                   0x7
   #define SUBCMD_DEVICE_SET_WATCHDOG                 0x8
   #define SUBCMD_DEVICE_GET_ARINFO                   0x9
   #define SUBCMD_DEVICE_SETUP_FIXED_MODULES          0xA

#define CMD_DEV_OFFSET_TABLE                       0xE
   #define SUBCMD_GET_DEV_OFFSET_TABLE_VAL_COUNT      0x1
   #define SUBCMD_GET_DEV_OFFSET_TABLE                0x2
   #define SUBCMD_GET_DEV_PROCESSIMAGELEN             0x3
   #define SUBCMD_GET_DEV_REMAP_TABLE                 0x4
   #define SUBCMD_DEV_ENABLE_REMAPPING                0x5

//This Command / Subcommands are handled internally by the dfpnio.dll The Controller card has no knowledge about them.
#define CMD_PN_DEV_CYCLIC_DATA_TRANSFER            0xF
   #define SUBCMD_PN_DEV_SET_GET_DATA                 0x1//   // cyclic complete process image data exchange
   #define SUBCMD_PN_DEV_GET_OUTPUT_DATA              0x2//   // cyclic complete output data receipt
   #define SUBCMD_PN_DEV_SET_INPUT_DATA               0x3//   // cyclic complete input data sending
   #define SUBCMD_PN_DEV_GET_INPUT_DATA               0x4//   // for access to parts of the process image input data
   #define SUBCMD_PN_DEV_GET_IOCS_STATUS              0x5//   // for access to parts of the process image output C status
   #define SUBCMD_PN_DEV_GET_IOPS_STATUS              0x6//   // for access to parts of the process image output P status
   #define SUBCMD_PN_DEV_SET_GET_DEVICE_DATA          0x7
   #define SUBCMD_PN_DEV_GET_STATUS                   0x8//   // Get only the status bytes


#define CMD_HW_INFO_DEV                            0x10
   // Same Subcommands as with CMD_HW_INFO


/****  Test Command DEFINES                                            ****/
#define CMD_TEST_CRIO                              0x82 // only one byte command
#define CMD_TEST                                   0x40000002
   #define SUBCMD_TPMANIP_ON                          0x1
   #define SUBCMD_TPMANIP_OFF                         0x2
   #define SUBCMD_TEST_DMA_COPY_TIMES                 0x3      //old define 0x40000001
   #define SUBCMD_TEST_BESTUECKERTEST                 0x4
   #define SUBCMD_TEST_BESTUECKERTEST_W_O_BUS_START_STOP    0x5
   #define SUBCMD_TEST_LIU                            0x6

#define CMD_UDP_USER                                0x11
    #define SUBCMD_CREATE_SOCKET                        0x1
    #define SUBCMD_DELETE_SOCKET                        0x2
    #define SUBCMD_SEND_DATA                            0x3
    #define SUBCMD_RECEIVE_DATA                         0x4


/*****************************************************************************
 **** DEFINES Common                                                      ****
 *****************************************************************************/
#define DCP_STD_TIMEOUT_MS                         10000
#define RECORD_STD_TIMEOUT_S                       10


/*****************************************************************************
 **** DEFINES PN operation modes                                          ****
 *****************************************************************************/
#define HW_DF_PN_MODE_CONTROLLER       0           // default
#define HW_DF_PN_MODE_DEVICE           1
#define HW_DF_PN_MODE_DEVICE_SIMULAT   2


/*****************************************************************************
**** TYPDEFS Commands/Subcommands                                         ****
*****************************************************************************/

/*****************************************************************************
**** TYPDEFS CMD_PN_DCP                                                   ****
*****************************************************************************/

typedef struct TAG_DCP_RESULT_TYPE
{
    INT32U    XID;                                /* Transaction identifier             */
    INT8U    ServiceID;                          /* This variable contains the DCP service number
                                                     which issued this confirmation.
                                                     Allowed values: DCP_GET, DCP_SET, DCP_IDENTIFY*/
    INT8U    MACAddr[MAC_ADDR_SIZE];             /* This variable receives the MAC address of the DCP request target.*/
    INT8U    IPAddr[IP_ADDR_SIZE];               /* This attribute receives the IP address of the DCP request target.*/
    INT8U    Gateway[IP_ADDR_SIZE];              /* This attribute receives the standard gateway address of the DCP request target.*/
    INT8U    NetworkMask[IP_ADDR_SIZE];          /* This attribute receives the network mask of the DCP request target.*/
    INT8U    StationNameLength;                  /* This attribute receives the length of the station name.*/
    CHAR8             StationName[MAX_STATION_NAME_SIZE]; /* This attribute receives a pointer to a buffer with size MAX_STATION_NAME_SIZE
                                                     containing the station name.*/
    INT8U    AliasNameLength;                    /* This attribute receives the length of the alias name.*/
    CHAR8             AliasName[MAX_STATION_NAME_SIZE];   /* This attribute receives a pointer to a buffer with size MAX_STATION_NAME_SIZE
                                                     containing the alias name.*/
    INT8U    DevTypeLength;                      /* This attribute receives the length of the device type.*/
    CHAR8             DevType[MAX_STATION_NAME_SIZE];     /* This attribute receives a pointer to a buffer with size MAX_STATION_NAME_SIZE
                                                     containing the device type.*/
    INT16U   VendorID;                            /* This attribute receives the vendor ID.*/
    INT16U   DeviceID;                            /* This attribute receives the device ID.*/
    INT8U    DeviceRole;                          /* This attribute receives the device role.*/
    INT8U    Result;                              /* This variable contains the result of the DCP operation.*/
} DCP_RESULT_TYPE;

typedef struct TAG_DCP_ARP_TARGET_INFO_OLD
{
    INT8U      pbyMACAddr[MAC_ADDR_SIZE];
    INT8U      pbyIPAddr[IP_ADDR_SIZE];
    INT8U      pbySubnetMask[IP_ADDR_SIZE];
    INT8U      pbyStdGateway[IP_ADDR_SIZE];
    TBOOL    wBlockQualifier;
    INT8U      byNameLength;
    CHAR8        pcName[MAX_STATION_NAME_SIZE];
} T_DCP_ARP_TARGET_INFO_OLD;


/*! This structure contains information about the target in case of a DCP request.
    The elements must be filled with regard to service that shall be sent. */
typedef struct TAG_DCP_ARP_TARGET_INFO_TYPE
{
   INT8U      pbyMACAddr[MAC_ADDR_SIZE];
   INT8U      pbyIPAddr[IP_ADDR_SIZE];
   INT8U      pbySubnetMask[IP_ADDR_SIZE];
   INT8U      pbyStdGateway[IP_ADDR_SIZE];
   INT16U     wBlockQualifier;
   INT8U      byNameLength;
   CHAR8      pcName[MAX_STATION_NAME_SIZE];
} DCP_ARP_TARGET_INFO_TYPE;

typedef struct TAG_SUBCMD_PN_DCP_SEND_IDENTIFY_ALL_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMS;
   INT32U DcpSearchResultEntrys;
   DCP_RESULT_TYPE DcpSearchResultList[MAX_IDENTIFY_ALL_DEVICES_COUNT];
} SUBCMD_PN_DCP_SEND_IDENTIFY_ALL_TYPE;

typedef struct TAG_SUBCMD_PN_DCP_SEND_SET_NAME_IP_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMS;
   DCP_ARP_TARGET_INFO_TYPE DeviceParameter;
} SUBCMD_PN_DCP_SEND_SET_NAME_IP_TYPE;

typedef struct TAG_SUBCMD_PN_DCP_SEND_SET_NAME_IP_TYPE_OLD
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMS;
   T_DCP_ARP_TARGET_INFO_OLD DeviceParameter;
} SUBCMD_PN_DCP_SEND_SET_NAME_IP_TYPE_OLD;

typedef struct TAG_SUBCMD_PN_DCP_SEND_IDENTIFY_DEVICEPROP
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMS;
   INT16U VendorID;
   INT16U DeviceID;
   INT32U DcpSearchResultEntrys;
   DCP_RESULT_TYPE DcpSearchResultList[MAX_IDENTIFY_ALL_DEVICES_COUNT];
} SUBCMD_PN_DCP_SEND_IDENTIFY_DEVICEPROP_TYPE;

typedef struct TAG_SUBCMD_PN_DCP_SEND_CONTROL_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMS;
   INT32U DCP_CONTROL_COMMAND;
   DCP_ARP_TARGET_INFO_TYPE DeviceParameter;
} SUBCMD_PN_DCP_SEND_CONTROL_TYPE;

typedef struct TAG_SUBCMD_PN_DCP_SEND_CONTROL_TYPE_OLD
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMS;
   INT32U DCP_CONTROL_COMMAND;
   T_DCP_ARP_TARGET_INFO_OLD DeviceParameter;
} SUBCMD_PN_DCP_SEND_CONTROL_TYPE_OLD;

/*****************************************************************************
**** TYPDEFS CMD_ALARM                                                    ****
*****************************************************************************/

/*! This structure contains the Alarm Notification Parameter. */
typedef struct TAG_ALARM_NOTIFICATION_TYPE{
    INT16U  wBlockType;
    INT16U  wBlockVersion;
    INT16U  wBlockLength;
    INT16U  wAlarmType;
    INT32U  dwAPI;
    INT16U  wSlotNumber;
    INT16U  wSubslotNumber;
    INT32U  dwModuleIdentNumber;
    INT32U  dwSubmoduleIdentNumber;
    INT16U  wAlarmSpecifier;
    INT16U  wLengthOfAlarmItem;
    INT32U  Dummy; // for backward compatibility (*pbyAlarmItem can be 32 or 64 bit)
} ALARM_NOTIFICATION_TYPE;

//Currently we defined a maximum length of 1500. This is longer than a standard IP packet,
//and we assume that it will never be longer, due all profinet packets are limited to this lenght.
//BS 02.08.2013
#define MAX_ALARM_LEN_PER_DEVICE    1500
typedef struct TAG_SUBCMD_ALARM_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U AlarmMode;                              //currently unused. Will be used to switch between automatic Alarm acknowledge, and manual acknowledge
      #define AUTOMATIC_ALARM_ACKNOWLEDGE          0x1   //Acknowledge after receive
      #define MANUAL_ALARM_ACKNOWLEDGE             0x2   //Acknowledge with acyclic request
      #define READ_ALARM_ACKNOWLEDGE               0x3   //Acknowledge after alarm read
   INT32U AlarmClass;                             //The user has to request an alarm type. Valid values are PROCESS_ALARM or DIAGNOSIS_ALARM
   INT32U DeviceNO;                               //ARUID = AR_USER_ID = Device number from the Configurator II.  Valid values are < MAX_DEVICES_COUNT
   ALARM_NOTIFICATION_TYPE AlarmNot;              //Standard Alarm content, containing the length of the data following.
                                                         //The last element of the struct (pointer) is unsed and always zero. The data is following in the
                                                         //next element AlarmData
   INT8U AlarmData[MAX_ALARM_LEN_PER_DEVICE];    //Alarm content. Usually very short.
} SUBCMD_ALARM_TYPE;

/*****************************************************************************
**** TYPDEFS CMD_PN_RW_RECORD                                             ****
*****************************************************************************/

typedef struct {
   INT32U  DeviceNo;
   INT16U SlotNumber;
   INT16U SubslotNumber;
   INT16U Index;
   INT16U Reserved;
   INT32U  TransactionNumber;
   INT32U  DataLength;
} SUBCMD_PN_RW_RECORD_PARAM_IN_TYPE;

typedef struct {
   INT8U  ErrorCode;
   INT8U  ErrorDecode;
   INT8U  ErrorCode1;
   INT8U  ErrorCode2;
   INT16U AddData1;
   INT16U AddData2;
   INT32U  DataLength;
} SUBCMD_PN_RW_RECORD_READ_CONF_PARAM_TYPE;

typedef struct TAG_SUBCMD_PN_RW_RECORD_READ_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U timeoutSec;
   SUBCMD_PN_RW_RECORD_PARAM_IN_TYPE ReadRecord;
   SUBCMD_PN_RW_RECORD_READ_CONF_PARAM_TYPE ReadRetRecord;
   INT8U RecordData[MAX_SIZE_OF_INTERFACE_CHANNEL - INTERFACE_STD_HEADER_TYPE_SIZE - 4 - sizeof(SUBCMD_PN_RW_RECORD_PARAM_IN_TYPE) - sizeof(SUBCMD_PN_RW_RECORD_READ_CONF_PARAM_TYPE)];
} SUBCMD_PN_RW_RECORD_READ_TYPE;
#define SUBCMD_PN_RW_RECORD_READ_TYPE_SIZE sizeof(SUBCMD_PN_RW_RECORD_READ_TYPE)

typedef struct {
   INT8U  ErrorCode;
   INT8U  ErrorDecode;
   INT8U  ErrorCode1;
   INT8U  ErrorCode2;
   INT16U AddData1;
   INT16U AddData2;
} SUBCMD_PN_RW_RECORD_WRITE_CONF_PARAM_TYPE;

typedef struct TAG_SUBCMD_PN_RW_RECORD_WRITE_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U timeoutSec;
   SUBCMD_PN_RW_RECORD_PARAM_IN_TYPE  WriteRecord;
   SUBCMD_PN_RW_RECORD_WRITE_CONF_PARAM_TYPE WriteRetRecord;
   INT8U RecordData[MAX_SIZE_OF_INTERFACE_CHANNEL - INTERFACE_STD_HEADER_TYPE_SIZE - 4 - sizeof(SUBCMD_PN_RW_RECORD_PARAM_IN_TYPE) - sizeof(SUBCMD_PN_RW_RECORD_WRITE_CONF_PARAM_TYPE)];
} SUBCMD_PN_RW_RECORD_WRITE_TYPE;
#define SUBCMD_PN_RW_RECORD_WRITE_TYPE_SIZE sizeof(SUBCMD_PN_RW_RECORD_WRITE_TYPE)

/*****************************************************************************
**** TYPDEFS CMD_PN_CYCLIC_DATA_TRANSFER                                  ****
*****************************************************************************/
typedef struct TAG_SUBCMD_PN_SET_GET_DEVICE_DATA_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT16U DeviceNo;
   INT8U APDUStatus;
   INT8U DeviceStatus;
   INT8U InputStatus;
   INT8U Padding1;
   INT16U InputLength;
   INT8U InputData[0x1000];
   INT8U OutputStatus;
   INT8U Padding2;
   INT16U OutputLength;
   INT8U OutputData[0x1000];
} SUBCMD_PN_SET_GET_DEVICE_DATA_TYPE;
#define SUBCMD_PN_SET_GET_DEVICE_DATA_TYPE_SIZE sizeof(SUBCMD_PN_SET_GET_DEVICE_DATA_TYPE)


/*****************************************************************************
**** TYPDEFS CMD_OFFSET_TABLE                                             ****
*****************************************************************************/
/*  1 MB - length of standard header - length  OffsetTableValCount */
#define MAX_SIZE_OF_OFFSET_TABLE       (MAX_SIZE_OF_INTERFACE_CHANNEL - INTERFACE_STD_HEADER_TYPE_SIZE - sizeof(INT32U))
#define MAX_NUM_OF_INT_OF_OFFSET_TABLE (MAX_SIZE_OF_OFFSET_TABLE/4)  //  /4 because of type INT32U
typedef struct TAG_CMD_OFFSET_TABLE_RET_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U OffsetTableValCount;
   INT32U OffsetTable[MAX_NUM_OF_INT_OF_OFFSET_TABLE];
} CMD_OFFSET_TABLE_RET_TYPE;

typedef struct TAG_CMD_OFFSET_TABLE_LEN_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U OffsetTableValCount;
} CMD_OFFSET_TABLE_LEN_TYPE;
#define CMD_OFFSET_TABLE_LEN_TYPE_SIZE sizeof(CMD_OFFSET_TABLE_LEN_TYPE)

typedef struct TAG_CMD_OFFSET_TABLE_PROCESSIMAGELEN_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U InputDataProcessImageLen;
   INT32U OutputDataProcessImageLen;
} CMD_OFFSET_TABLE_PROCESSIMAGELEN_TYPE;
#define CMD_OFFSET_TABLE_PROCESSIMAGELEN_TYPE_SIZE sizeof(CMD_OFFSET_TABLE_PROCESSIMAGELEN_TYPE)

/* Data remapping:
 *
 * Process Image  | Device 0 | Device 1 | Device 2 | .....
 *                     ||
 *                     \/
 *                 -------------------------------------------------
 * Device Image   | APDU  | Device Status | IOPS | IOCS | Data ...  |
 *
 */
typedef struct TAG_PD_REMAP_HEADER_TYPE
{
    INT8U APDUStatus;
    INT8U DeviceStatus;
    INT8U IOPS;
    INT8U IOCS;
    // ..... Data
}PD_REMAP_HEADER_TYPE;

typedef struct TAG_CMD_REMAP_DEVICE_TYPE
{
    INT32U ARUID;
    INT32U InputOffset;  // Begin of Device in Process Image
    INT32U InputLength;  // Process Data Length
    INT32U OutputOffset; // Begin of Device in Process Image
    INT32U OutputLength; // Process Data Length
} CMD_REMAP_DEVICE_TYPE;

typedef struct TAG_CMD_REMAP_OFFSET_TABLE_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U NumberOfDevices;
   INT8U Reserved[28]; // Cycle Time, etc ?
   CMD_REMAP_DEVICE_TYPE OffsetTable[MAX_DEVICES_COUNT];
} CMD_REMAP_OFFSET_TABLE_TYPE;
#define CMD_REMAP_OFFSET_TABLE_TYPE_SIZE sizeof(CMD_REMAP_OFFSET_TABLE_TYPE)


typedef struct TAG_CMD_REMAP_ENABLE_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT8U Enable;
} CMD_REMAP_ENABLE_TYPE;

/*****************************************************************************
**** TYPDEFS CMD_HW_INFO                                                  ****
*****************************************************************************/
typedef struct TAG_HW_PARAMETER_PN_IO_REQ_TYP
{
   //Format       Entry name                 Length   Position    UdpScan?    Mand. UDP?     Description
   INT32U   IpAddress;//               //  4      0         UDP:yes     mandatory
   INT32U   SubNetMask;//              //  4      4         UDP:yes     mandatory
   INT32U   GateWay;//                 //  4      8         UDP:yes     mandatory
   INT8U    MacAddress[8];//           //  8     12         UDP:yes     mandatory      Used on all devices for the Main Mac Adress
   INT32U   SerialNumber;//            //  4     20         UDP:yes     mandatory      Used on all devices for the Main Mac Adress
   INT32U   HWType;//                  //  4     24         UDP:yes     mandatory
   INT32U   DHCPActive;//              //  4     28         UDP:yes     mandatory      0 ==.. inkativ    // 0 !=.. aktiv
   INT32U   AliveTimeout;//            //  4     32         UDP:yes     no             Only used from Profibus based Products FNL/DF PROFI II/Proxy
   INT32U   DebugLevel;//              //  4     36         UDP:yes     no             Sensefull for all products. Available on allProfibus/Profinet based products
   INT32U   PNVendorID;//              //  4     40         UDP:yes     no             Only Profinet products
   INT32U   PNDeviceID;//              //  4     44         UDP:yes     no             Only Profinet products   INT16U CS_SW_Ver_First;//         // 2     8                       UDP:yes     mandatory      Major Versionnumber of the embedded firmware
   INT16U   CS_SW_Ver_First;//         //  2     48         UDP:yes     mandatory      Major Versionnumber of the profinet SW module
   INT16U   CS_SW_Ver_Second;//        //  2     50         UDP:yes     mandatory      Middle Versionnumber of the profinet SW module
   INT32U   CS_SW_Ver_Third;//         //  4     52         UDP:yes     no             Minor Versionnumber of the profinet SW module
   INT8U    VersionString[32];//       // 32     56         UDP:yes     mandatory      free writeable Version string
   INT32U   HardwareRevision;//        //  4     88         UDP:yes     no             Hardware Revison
   INT16S   Temperature_1;//           //  4     92         UDP:yes     no             Temperatur results from CBIT for chip near to processor
   INT16S   Temperature_2;//           //  4     94         UDP:yes     no             Temperatur results from CBIT for chip in the upper left corner nexto to LEDs
   INT32U   PNBusRunning;//            //  4     96         UDP:yes     no             Indicates if Profinet is active (0 = Profinet not active, otherwise active)
   INT8U    CS_OS_Ver_First;//         //  1    100         UDP:yes     no             Major Versionnumber of the OS
   INT8U    CS_OS_Ver_Second;//        //  1    101         UDP:yes     no             Middle Versionnumber of the OS
   INT8U    CS_OS_Ver_Third;//         //  1    102         UDP:yes     no             Minor Versionnumber of the OS
   INT8U    Padding1;//                //  1    103         UDP:no      no             Might be replaced by future values.
   INT32U   DriverType;//              //  4    104         UDP:yes     no             Driver Type
   INT32U   DriverVersion1;//          //  4    108         UDP:yes     no             Driver Version 1 highest significant version number
   INT32U   DriverVersion2;//          //  4    112         UDP:yes     no             Driver Version 2
   INT32U   DriverVersion3;//          //  4    116         UDP:yes     no             Driver Version 3
   INT32U   DriverVersion4;//          //  4    120         UDP:yes     no             Driver Version 4 lowest significant version number
   INT32U   TCPTerminal;//             //  4    124         UDP:yes     no             Enabled TCP Terminal
   INT8U    ReservedForExt[128];//     //116    140         UDP:no      no             For Extension in future
   //                                //Sum: 256 = 0x100
} HW_PARAMETER_PN_IO_REQ_TYP;
#define HW_PARAMETER_PN_IO_REQ_TYP_LEN        sizeof(HW_PARAMETER_PN_IO_REQ_TYP)

typedef struct TAG_UDP_COM_STRUCT
{
   INT8U CommandString[20]; //DISCOVER_STRING or RESPONSE_STRING or CONFIG_STRING
      #define  DISCOVER_STRING "CS DEVICE DISCOVERY"
      #define  RESPONSE_STRING "CS DEVICE RESPONSE "
      #define  CONFIG_STRING   "CS DEVICE CONFIG   "
   INT8U DeviceString[16];
      #define  DEVICE_STRING_ALL          "CS_DEVICES     "
      #define  DEVICE_STRING_DF_PN_IO     "DF PN IO       "
   HW_PARAMETER_PN_IO_REQ_TYP SerialNumbRequestStruct;
} UDP_PN_IO_COM_STRUCT;

typedef struct TAG_SUBCMD_HW_INFO_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   UDP_PN_IO_COM_STRUCT UDP_Com;
   INT32U InterfaceIndex;
} SUBCMD_HW_INFO_TYPE;

typedef struct TAG_SUBCMD_HW_INFO_SET_DEBUG_LEVEL_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U DebugLevel;
} SUBCMD_HW_INFO_SET_DEBUG_LEVEL_TYPE;

typedef struct TAG_SUBCMD_HW_INFO_SET_TIME_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32S Time_in_Seconds;                //Absolute time based on the Unix epoch 00:00:00 UTC off 1. January 1970.
   INT32S Time_in_NanoSeconds;            //number of passed nanoseconds inside the specified seconde of the Unix epoch. max: 1000 000 000 = 0x3B9A CA00
} SUBCMD_HW_INFO_SET_TIME_TYPE;

typedef struct TAG_SUBCMD_HW_INFO_SET_STATIONNAME_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   CHAR8 StationName[240*2+2];              // 240 Byte string len + termination
} SUBCMD_HW_INFO_SET_STATIONNAME_TYPE;

/*****************************************************************************
**** TYPDEFS CMD_FILE_TRANSFER                                            ****
*****************************************************************************/
#define CMD_FILE_TRANSFER_TYPE_COMMAND_OVERHEAD   (4*4 + 48)
typedef struct TAG_CMD_FILE_TRANSFER_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U PacketNo;
   INT32U PacketCount;
   INT32U FileLocation;
      #define  RAM_DRIVE                  1
      #define  FLASH_DRIVE                2
      #define  U_BOOT_IMAGE               3
      #define  VXWORKS_BOOT_LOADER_IMAGE  4
      #define  FLASH_IMAGE                5
      #define  BOOT_LINE                  6
   INT32U CompressionMode;
      #define  NO_COMPRESSION             1
      #define  COMPRESS_FILE              2
      #define  DECOMPRESS_FILE            3
   INT8U FileName[48];
   INT8U Data[MAX_SIZE_OF_INTERFACE_CHANNEL - INTERFACE_STD_HEADER_TYPE_SIZE - CMD_FILE_TRANSFER_TYPE_COMMAND_OVERHEAD];
} CMD_FILE_TRANSFER_TYPE;

/*****************************************************************************
**** TYPDEFS CMD_VISA_DLL                                                 ****
*****************************************************************************/
typedef struct TAG_SUBCMD_VISA_DLL_GET_ALARM_TEXT_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U DeviceNo;
   INT32U AlarmClass;
   INT8U AlarmText[0x1000];
} SUBCMD_VISA_DLL_GET_ALARM_TEXT_TYPE;

typedef struct TAG_SUBCMD_VISA_DLL_GET_ERROR_TEXT_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT8U ErrorText[0x1000];
} SUBCMD_VISA_DLL_GET_ERROR_TEXT_TYPE;

typedef struct TAG_SUBCMD_VISA_DLL_GET_MODULEDIFFBLOCK_TEXT_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U DeviceNo;
   INT8U ModuleDiffBlockText[0x1000];
} SUBCMD_VISA_DLL_GET_MODULEDIFFBLOCK_TEXT_TYPE;

/*****************************************************************************
**** TYPDEFS CMD_DIAG                                                     ****
*****************************************************************************/

/*! This structure comprises the output parameters of the AR. */
typedef struct TAG_AR_READ_PARAM_OUT_TYPE{
    INT8U            byIODeviceStationNameLen;
    INT32U           dummyIODeviceStationName; // comptability pointer
    INT16U           wARID;
    INT16U           wARUserID;
    INT16U           wARStatus;
    INT16U           wARAddInfo;
    INT8U            byErrorCode;
    INT8U            byErrorDecode;
    INT8U            byErrorCode1;
    INT8U            byErrorCode2;
    INT32U           dummyARwithStatusList; // comptability pointer
} AR_READ_PARAM_OUT_TYPE;

typedef struct TAG_SUBCMD_DIAG_DEVICE_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U DeviceNO;
   INT32U ARDiag;
   AR_READ_PARAM_OUT_TYPE ARParam;
} SUBCMD_DIAG_DEVICE_TYPE;

typedef struct TAG_MODULE_DIFF_BLOCK_SUBSLOT_TYPE
{
   INT32U SubslotNumber;
   INT32U SubmoduleIdentNumber;
   INT32U SubmoduleState;
} MODULE_DIFF_BLOCK_SUBSLOT_TYPE, *P_MODULE_DIFF_BLOCK_SUBSLOT_TYPE;

typedef struct TAG_MODULE_DIFF_BLOCK_SLOT_TYPE
{
   INT32U SlotNumber;
   INT32U ModuleIdentNumber;
   INT32U ModuleState;
   INT32U NumberOfSubmodules;
      #define MODULE_DIFF_BLOCK_MAX_NUMBER_OF_SUBMODULES 25
   MODULE_DIFF_BLOCK_SUBSLOT_TYPE ListOfSubmodules[MODULE_DIFF_BLOCK_MAX_NUMBER_OF_SUBMODULES];
} MODULE_DIFF_BLOCK_SLOT_TYPE, *P_MODULE_DIFF_BLOCK_SLOT_TYPE;

typedef struct TAG_MODULE_DIFF_BLOCK_TYPE
{
   INT32U ARID;
   CHAR8 StationName[MAX_STATION_NAME_SIZE];
   INT32U NumberOfModules;
      #define MODULE_DIFF_BLOCK_MAX_NUMBER_OF_MODULES    25
   MODULE_DIFF_BLOCK_SLOT_TYPE ListOfModules[MODULE_DIFF_BLOCK_MAX_NUMBER_OF_MODULES];
} MODULE_DIFF_BLOCK_TYPE, *P_MODULE_DIFF_BLOCK_TYPE;

typedef struct TAG_SUBCMD_DIAG_MODULE_DIFF_BLOCK
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U DeviceNO;
   MODULE_DIFF_BLOCK_TYPE ModuleDiffBlock;
} SUBCMD_DIAG_MODULE_DIFF_BLOCK_TYPE;
#define SUBCMD_DIAG_MODULE_DIFF_BLOCK_TYPE_SIZE sizeof(SUBCMD_DIAG_MODULE_DIFF_BLOCK_TYPE)


/*****************************************************************************
**** TYPDEFS CMD_DEVICE                                                   ****
*****************************************************************************/
typedef struct TAG_AR_INFO_TYPE
{
   INT8U  CMInitName[MAX_STATION_NAME_SIZE];  /*!< Station name */
   INT8U  CMInitIPAddr[IP_ADDR_SIZE];         /*!< IP address */
   INT8U  CMInitMacAddr[MAC_ADDR_SIZE];       /*!< MAC address (Device only) */
   INT8U  DAPConfigured;                      /*!< TRUE if this AR has a configured DAP (Device only) */
   INT8U  Reserved8;
   INT16U ARID;                               /*!< Inernal ARID */
   INT16U ARUID;                              /*!< DeviceNo to access the device */
   INT16U TxSendClockFactor;                  /*!< Sendclock Factor for tx direction */
   INT16U RxSendClockFactor;                  /*!< Sendclock Factor for rx direction */
   INT16U TxReductionRatio;                   /*!< Reduction Ratio for tx direction */
   INT16U RxReductionRatio;                   /*!< Reduction Ratio for rx direction */
   INT16U MaxAlarmDataLength;                 /*!< Maximum Alarm Data Length */
   INT16U Reserved16;
} AR_INFO_TYPE, *P_AR_INFO_TYPE;

typedef struct TAG_SUBCMD_GET_ARINFO_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT16U NumberOfAr;
   INT16U Reserved;
   AR_INFO_TYPE  Ar[MAX_DEVICES_COUNT];

} SUBCMD_DEVICE_GET_ARINFO_TYPE, *P_SUBCMD_DEVICE_GET_ARINFO_TYPE;
#define SUBCMD_DEVICE_GET_ARINFO_TYPE_SIZE sizeof(SUBCMD_DEVICE_GET_ARINFO_TYPE)

typedef struct TAG_SUBCMD_DEVICE_SETUP_FIXED_MODULES_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT8U Enable;
   INT8U Reserved[3];
   INT32U ModuleID[MAX_NUM_OF_SLOT];
} SUBCMD_DEVICE_SETUP_FIXED_MODULES_TYPE, *P_SUBCMD_DEVICE_SETUP_FIXED_MODULES_TYPE;
#define SUBCMD_DEVICE_SETUP_FIXED_MODULES_TYPE_SIZE sizeof(SUBCMD_DEVICE_SETUP_FIXED_MODULES_TYPE)

/*! This structure contains the description of a diagnosis entry. */
typedef struct TAG_PNDIAG_DIAG_INFO_TYPE
{
   INT16U      wChannelProperties;
   INT16U      wUserStructureIdentifier;
   INT16U      wChannelErrorType;
   INT16U      wExtChannelErrorType;
   INT32U      dwExtChannelAddValue;
   INT32U      dwQualifiedChannelQualifier;
   INT8U       byRTATrigger;
} PNDIAG_DIAG_INFO_TYPE;

typedef struct TAG_SUBCMD_DEVICE_DIAG_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT16U Slot;
   INT16U Channel;
   PNDIAG_DIAG_INFO_TYPE  PnDiagItem;
} SUBCMD_DEVICE_DIAG_TYPE, *P_SUBCMD_DEVICE_DIAG_TYPE;
#define SUBCMD_DEVICE_DIAG_TYPE_SIZE sizeof(SUBCMD_DEVICE_DIAG_TYPE)

typedef struct TAG_DEVICE_DIAG_TYPE
{
   INT16U UserStructureIdentifier;
   INT16U ChannelNumber;
   INT16U ChannelProperties;
   INT16U ChannelErrorType;
   INT16U ExtChannelErrorType;
   INT32U  ExtChannelAddValue;
   INT32U  QualifiedChannelQualifier;
} DEVICE_DIAG_TYPE;
#define DEVICE_DIAG_TYPE_SIZE sizeof(DEVICE_DIAG_TYPE)

typedef struct TAG_SUBCMD_DEVICE_SEND_ALARM_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT16U Slot;
   INT16U SubSlot;
   INT16U AlarmType;
   INT16U ItemLength;
   union
   {
      DEVICE_DIAG_TYPE  DiagItem;
      INT8U  ManufacturerItemU8[200];
      INT16U ManufacturerItemU16[100];
      INT32U  ManufacturerItemU32[50];
   } Item; // optional
} SUBCMD_DEVICE_SEND_ALARM_TYPE, *P_SUBCMD_DEVICE_SEND_ALARM_TYPE;
#define SUBCMD_DEVICE_SEND_ALARM_TYPE_SIZE sizeof(SUBCMD_DEVICE_SEND_ALARM_TYPE)

typedef struct TAG_SUBCMD_DEVICE_MOD_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT16U Slot;
   INT16U SubSlot;

#define CHANGE_MOD_PULL          0
#define CHANGE_MOD_PLUG          1
   INT16U Action;
} SUBCMD_DEVICE_MOD_TYPE, *P_SUBCMD_DEVICE_MOD_TYPE;
#define SUBCMD_DEVICE_MOD_TYPE_SIZE sizeof(SUBCMD_DEVICE_MOD_TYPE)

typedef struct TAG_DEVICE_RECORD_TYPE
{
   INT8U WriteAccess;
   INT8U ReadAccess;
   INT16U Length;
   INT8U Data[MAX_DEVICE_RECORD_SIZE];
}DEVICE_RECORD_TYPE, *P_DEVICE_RECORD_TYPE;
#define DEVICE_RECORD_TYPE_SIZE sizeof(DEVICE_RECORD_TYPE)

typedef struct TAG_SUBCMD_DEVICE_RECORD_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT16U Slot;
   INT16U SubSlot;
   INT16U RecordIndex; // starting with 0
   DEVICE_RECORD_TYPE Record;
}SUBCMD_DEVICE_RECORD_TYPE, *P_SUBCMD_DEVICE_RECORD_TYPE;
#define SUBCMD_DEVICE_RECORD_TYPE_SIZE sizeof(SUBCMD_DEVICE_RECORD_TYPE)

typedef struct TAG_SUBCMD_DEVICE_SET_WATCHDOG_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TimeoutMs;
}SUBCMD_DEVICE_SET_WATCHDOG_TYPE, *P_SUBCMD_DEVICE_SET_WATCHDOG_TYPE;
#define SUBCMD_DEVICE_SET_WATCHDOG_TYPE_SIZE sizeof(SUBCMD_DEVICE_SET_WATCHDOG_TYPE)



/*****************************************************************************
**** TYPDEFS CMD_TEST                                                     ****
*****************************************************************************/
typedef struct TAG_SUBCMD_TPMANIP_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U TracePointId;
} SUBCMD_TPMANIP_TYPE;

typedef struct TAG_SUBCMD_TEST_LIU_TYPE
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   INT32U LiuCommand;
   INT32U opt1;
   INT32U opt2;
   INT32U opt3;
   INT32U opt4;
   INT32U opt5;
} SUBCMD_TEST_LIU_TYPE;


/*****************************************************************************
**** TYPDEFS UDP_USER                                                     ****
*****************************************************************************/
#define UDP_USER_MAX_NUM_SOCKETS    10
#define UDP_TELEGRAM_LEN            1480 // max Payload in Bytes of one PDU Telegram

typedef struct TAG_CMD_UDP_USER_TYPE
{
    INTERFACE_STD_HEADER_TYPE StdHeader;
    INT8S   s8UserSocketIdx;    // reference to array where socket is stored
    INT32U  u32IpAddress;
    INT16U  u16Port;
    INT8U   ac8MsgData[UDP_TELEGRAM_LEN];
    INT32U  u32MsgLen;
    INT8U   u8UdpDataAvailable; // > 0 = more Data available to read
                                //       - User should call this function with same ps8UserSockIdx again
} CMD_UDP_USER_TYPE;

/*!
 * Union with all available communication structures.
 * This can be used as generic type for communication functions.
 */
typedef union TAG_COMMON_INTERFACE_UNION
{
   INTERFACE_STD_HEADER_TYPE StdHeader;
   COMMON_INTERFACE_TYPE CommonInterfaceType;
   SUBCMD_PN_DCP_SEND_IDENTIFY_ALL_TYPE DcpIdentAll;
   SUBCMD_PN_DCP_SEND_SET_NAME_IP_TYPE DcpSetNameIP;
   SUBCMD_PN_DCP_SEND_SET_NAME_IP_TYPE_OLD DcpSetNameIP_old;
   SUBCMD_PN_DCP_SEND_IDENTIFY_DEVICEPROP_TYPE DcpIdentDeviceProp;
   SUBCMD_PN_DCP_SEND_CONTROL_TYPE DcpControl;
   SUBCMD_PN_DCP_SEND_CONTROL_TYPE_OLD DcpControl_old;
   SUBCMD_ALARM_TYPE Alarm;
   SUBCMD_PN_RW_RECORD_WRITE_TYPE WriteRecord;
   SUBCMD_PN_RW_RECORD_READ_TYPE ReadRecord;
   CMD_OFFSET_TABLE_LEN_TYPE OffsetTableLen;
   CMD_OFFSET_TABLE_RET_TYPE OffsetTableRet;
   CMD_REMAP_OFFSET_TABLE_TYPE RemapOffsetTable;
   CMD_REMAP_ENABLE_TYPE RemapEnable;
   CMD_OFFSET_TABLE_PROCESSIMAGELEN_TYPE OffsetTableProcessImageLen;
   SUBCMD_HW_INFO_TYPE HW_Info;
   SUBCMD_HW_INFO_SET_DEBUG_LEVEL_TYPE HW_Info_Set_Debug_Level;
   SUBCMD_HW_INFO_SET_TIME_TYPE HW_Info_Set_Time;
   SUBCMD_HW_INFO_SET_STATIONNAME_TYPE HW_Info_Set_StationName;
   CMD_FILE_TRANSFER_TYPE File_Transfer;
   SUBCMD_DIAG_DEVICE_TYPE Diag_Device;
   SUBCMD_DIAG_MODULE_DIFF_BLOCK_TYPE Diag_Module_Diff_Block;
   SUBCMD_TPMANIP_TYPE TpManip;
   SUBCMD_TEST_LIU_TYPE Liu;
   SUBCMD_PN_SET_GET_DEVICE_DATA_TYPE getSetDeviceData;
   SUBCMD_DEVICE_SEND_ALARM_TYPE SendDeviceAlarm;
   SUBCMD_DEVICE_DIAG_TYPE DeviceDiagnosis;
   SUBCMD_DEVICE_MOD_TYPE DeviceChangeMod;
   SUBCMD_DEVICE_RECORD_TYPE DeviceRecord;
   SUBCMD_DEVICE_SET_WATCHDOG_TYPE AppWatchdog;
   SUBCMD_DEVICE_GET_ARINFO_TYPE DeviceArInfo;
   SUBCMD_DEVICE_SETUP_FIXED_MODULES_TYPE DeviceModulesFixed;
   CYCLIC_PROCESSIMAGE_TYPE ProcessImage;
   SUBCMD_VISA_DLL_GET_MODULEDIFFBLOCK_TEXT_TYPE VisaGetModuleDiffText;
   SUBCMD_VISA_DLL_GET_ERROR_TEXT_TYPE VisaGetErrorText;
   SUBCMD_VISA_DLL_GET_ALARM_TEXT_TYPE VisaGetAlarmText;
   CMD_UDP_USER_TYPE UdpUser;
} COMMON_INTERFACE_UNION;


/** @} */ // end of commifpn

#pragma pack(pop, Kunbus)

#endif // COMM_INTERFACE_H
